
import sys
import os
import logging.config
import string
import shutil
import socket
import time
import re
import stat
import shlex
import types
import fnmatch
import subprocess
from collections import OrderedDict
from traceback import format_exception
from threading import currentThread, RLock
from queue import Queue, Empty
from glob import glob
from datetime import datetime
from pickle import Unpickler, UnpicklingError
from locale import getpreferredencoding


class Callable:
    def __init__(self, method, *args):
        self.method = method
        self.extraArgs = args

    def __call__(self, *calledArgs, **kw):
        toUse = calledArgs + self.extraArgs
        return self.method(*toUse, **kw)

    def __eq__(self, other):
        return isinstance(other, Callable) and self.method == other.method and self.extraArgs == other.extraArgs

    def __hash__(self):
        return hash((self.method, self.extraArgs))

    def __deepcopy__(self, memo):
        return self  # don't copy these


def findInstallationRoots():
    packageDir = os.path.dirname(__file__)
    if getattr(sys, 'frozen', False):
        roots = glob(os.path.join(os.path.dirname(sys.executable), "lib", "python*", "site-packages", "texttestlib"))
    else:
        roots = [packageDir]
    installationRoot = os.path.dirname(packageDir)
    if os.path.basename(installationRoot) == "generic":
        siteRoot = os.path.dirname(installationRoot)
        roots.append(siteRoot)
    else:
        siteDir = os.path.join(installationRoot, "site")
        if os.path.isdir(siteDir):
            roots.append(siteDir)
    return roots


globalStartTime = datetime.now()
datetimeFormat = "%d%b%H:%M:%S"
installationRoots = findInstallationRoots()
# Don't read these from Python as the names depend on the locale!
weekdays = ["monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"]
# A few version control systems we might run into... list lifted from emacs "grep-find-ignored-directories"
# and removed the ones I'd never heard of...
controlDirNames = ["CVS", ".svn", ".bzr", ".hg", ".git", "RCS", "_darcs", "{arch}"]

# Useful utility...


def localtime(format=datetimeFormat):
    return datetime.now().strftime(format)


def startTimeString(format=datetimeFormat):
    return globalStartTime.strftime(format)


def isModuleMissing(errorString, moduleName):
    if not errorString.startswith("No module named"):
        return False

    moduleMissing = errorString.split()[-1][1:-1]
    return moduleName.startswith(moduleMissing)


def importAndCall(moduleName, callableName, *args):
    command = "from " + moduleName + " import " + callableName + " as _callable"
    namespace = {}
    try:
        exec(command, globals(), namespace)
    except ImportError as err:
        # try resolve import by prepending 'texttestlib.' (python3)
        if isModuleMissing(str(err), moduleName):
            moduleName = "texttestlib." + moduleName
            command = "from " + moduleName + " import " + callableName + " as _callable"
            try:
                exec(command, globals(), namespace)
            except ImportError as err2:
                if isModuleMissing(str(err2), moduleName):
                    raise err
                else:
                    raise err2
        else:
            raise

    return namespace["_callable"](*args)


def installationDir(name):
    # Generic modules only, we're confident we know where they are
    return os.path.join(installationRoots[0], name)


def getTextTestProgram():
    if getattr(sys, 'frozen', False):
        executable = sys.executable
    else:
        executable = os.path.abspath(sys.argv[0])
    slaveCmd = os.getenv("TEXTTEST_SLAVE_CMD", executable) # TextTest start-script for starting subsequent processes
    if not slaveCmd:
        # To allow it to be reset, the above form is for documentation...
        slaveCmd = executable
    return slaveCmd


def installationPath(*pathElems):
    for instRoot in installationRoots:
        instPath = os.path.join(instRoot, *pathElems)
        if os.path.exists(instPath):
            return instPath


def getPersonalDir(dataDirName):
    envVar = "TEXTTEST_PERSONAL_" + dataDirName.upper()
    return os.getenv(envVar, os.path.join(getPersonalConfigDir(), dataDirName))


def quote(value):
    quoteChar = "'"
    if quoteChar in value:
        return value  # don't double-quote
    # Make sure the home directory gets expanded...
    quotedValue = quoteChar + value + quoteChar
    quotedValue = quotedValue.replace("${", quoteChar + "${")
    quotedValue = quotedValue.replace("}", "}" + quoteChar)
    if quotedValue.startswith(quoteChar * 2):
        quotedValue = quotedValue[2:]
    if quotedValue.endswith(quoteChar * 2):
        quotedValue = quotedValue[:-2]
    return quotedValue


def pluralise(num, name):
    if num == 1:
        return "1 " + name
    else:
        return str(num) + " " + name + "s"


def getHideStartUpInfo():
    if os.name == "nt":
        info = subprocess.STARTUPINFO()
        # Python doesn't make this easy for us: in 2.6.6 and later these flags became inaccessible
        # Alternative is to use win32api which seems excessive just for this purpose.
        winFlagModule = subprocess if hasattr(subprocess, "STARTF_USESHOWWINDOW") else subprocess._subprocess  # @UndefinedVariable
        info.dwFlags |= winFlagModule.STARTF_USESHOWWINDOW
        info.wShowWindow = winFlagModule.SW_HIDE
        return info


def findDataDirs(includeSite=True, includePersonal=False, dataDirName="etc"):
    if includeSite:
        dirs = [os.path.join(instRoot, dataDirName) for instRoot in installationRoots]
    else:
        dirs = [os.path.join(installationRoots[0], dataDirName)]
    if includePersonal:
        dirs.append(getPersonalDir(dataDirName))
    return dirs


def findDataPaths(filePatterns, *args, **kwargs):
    paths = []
    for dir in findDataDirs(*args, **kwargs):
        for filePattern in filePatterns:
            paths += glob(os.path.join(dir, filePattern))
    return paths

# Parse a time string, either a HH:MM:SS string, or a single int/float,
# which is interpreted as a number of minutes, for backwards compatibility.
# Observe that in either 'field' in the HH:MM:SS case, any number is allowed,
# so e.g. 144:5.3:0.01 is interpreted as 144 hours + 5.3 minutes + 0.01 seconds.


def getNumberOfSeconds(timeString):
    parts = timeString.split(":")
    try:
        if len(parts) == 1:  # Backwards compatible, assume single ints/floats means minutes
            return 60 * float(timeString)
        elif len(parts) <= 3:                # Assume format is HH:MM:SS ...
            seconds = 0
            for i in range(len(parts) - 1, -1, -1):
                if (parts[i] != ""):  # Handle empty strings (<=:10 gives empty minutes field, for example)
                    seconds += float(parts[i]) * pow(60, len(parts) - 1 - i)
            return seconds
    except ValueError:
        pass
    raise TextTestError("Illegal time format '" + timeString +
                        "' :  Use format HH:MM:SS or MM:SS, or a single number to denote a number of minutes.")


def printWarning(message, stdout=False):
    if stdout:
        if log:
            log.warning("WARNING: " + message)
        else:
            print("WARNING: " + message)
    else:
        sys.stderr.write("WARNING: " + message + "\n")


# Useful stuff to handle regular expressions
regexChars = re.compile("[\^\$\[\]\{\}\\\*\?\|\+]")


def isRegularExpression(text):
    return (regexChars.search(text) != None)

# Useful for comparing strings with numbers in them
# so that e.g. "Group 9" comes before "Group 10"


def padNumbersWithZeroes(x):
    def padWithZeroes(match):
        return match.group(0).rjust(4, "0")
    return re.sub("[0-9]+", padWithZeroes, x)


# Parse the string as a byte expression.
# Mb/mb/megabytes/mbytes
def parseBytes(text):  # pragma: no cover
    lcText = text.lower()
    try:
        # Try this first, to save time if it works
        try:
            return float(text)
        except ValueError:
            pass

        if lcText.endswith("kb") or lcText.endswith("kbytes") or lcText.endswith("k") or lcText.endswith("kilobytes"):
            splitText = lcText.split('k')
            if len(splitText) > 2:
                raise
            return 1000 * float(splitText[0])
        elif lcText.endswith("kib") or lcText.endswith("kibibytes"):
            splitText = lcText.split('k')
            if len(splitText) > 2:
                raise
            return 1024 * float(splitText[0])
        elif lcText.endswith("mb") or lcText.endswith("mbytes") or lcText.endswith("m") or lcText.endswith("megabytes"):
            splitText = lcText.split('m')
            if len(splitText) > 2:
                raise
            return 1000000 * float(splitText[0])
        elif lcText.endswith("mib") or lcText.endswith("mebibytes"):
            splitText = lcText.split('m')
            if len(splitText) > 2:
                raise
            return 1048576 * float(splitText[0])
        elif lcText.endswith("gb") or lcText.endswith("gbytes") or lcText.endswith("g") or lcText.endswith("gigabytes"):
            splitText = lcText.split('g')
            if len(splitText) > 2:
                raise
            return 1000000000 * float(splitText[0])
        elif lcText.endswith("gib") or lcText.endswith("gibibytes"):
            splitText = lcText.split('g')
            if len(splitText) > 2:
                raise
            return 1073741824 * float(splitText[0])
        elif lcText.endswith("tb") or lcText.endswith("tbytes") or lcText.endswith("t") or lcText.endswith("terabytes"):
            splitText = lcText.split('t')
            if len(splitText) > 3:
                raise
            return 1000000000000 * float(splitText[0])
        elif lcText.endswith("tib") or lcText.endswith("tebibytes"):
            splitText = lcText.split('t')
            if len(splitText) > 3:
                raise
            return 1099511627776 * float(splitText[0])
        elif lcText.endswith("pb") or lcText.endswith("pbytes") or lcText.endswith("p") or lcText.endswith("petabytes"):
            splitText = lcText.split('p')
            if len(splitText) > 2:
                raise
            return 10**15 * float(splitText[0])
        elif lcText.endswith("pib") or lcText.endswith("pebibytes"):
            splitText = lcText.split('p')
            if len(splitText) > 2:
                raise
            return 2**50 * float(splitText[0])
        elif lcText.endswith("eb") or lcText.endswith("ebytes") or lcText.endswith("e") or lcText.endswith("exabytes"):
            splitText = lcText.split('e')
            if len(splitText) > 3:
                raise
            return 10**18 * float(splitText[0])
        elif lcText.endswith("eib") or lcText.endswith("exbibytes"):
            splitText = lcText.split('e')
            if len(splitText) > 3:
                raise
            return 2**60 * float(splitText[0])
        elif lcText.endswith("zb") or lcText.endswith("zbytes") or lcText.endswith("z") or lcText.endswith("zettabytes"):
            splitText = lcText.split('z')
            if len(splitText) > 2:
                raise
            return 10**21 * float(splitText[0])
        elif lcText.endswith("zib") or lcText.endswith("zebibytes"):
            splitText = lcText.split('z')
            if len(splitText) > 2:
                raise
            return 2**70 * float(splitText[0])
        elif lcText.endswith("yb") or lcText.endswith("ybytes") or lcText.endswith("y") or lcText.endswith("yottabytes"):
            splitText = lcText.split('y')
            if len(splitText) > 3:
                raise
            return 10**24 * float(splitText[0])
        elif lcText.endswith("yib") or lcText.endswith("yobibytes"):
            splitText = lcText.split('y')
            if len(splitText) > 2:
                raise
            return 2**80 * float(splitText[0])
        else:
            return float(text)
    except:
        raise "Illegal byte format '" + text + "'"

# pango markup doesn't like <,>,& ...


def convertForMarkup(message):
    return message.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")


class Filter:
    def acceptsTestCase(self, testArg):
        return 1  # pragma: no cover - implemented in all base classes

    def acceptsTestSuite(self, suiteArg):
        return 1

    def acceptsTestSuiteContents(self, suiteArg):
        return 1

    def refine(self, tests):
        return tests


class TextTriggerGroup:
    def __init__(self, texts):
        self.texts = texts
        self.textTriggers = [TextTrigger(text) for text in self.texts]

    def stringContainsText(self, searchString):
        for trigger in self.textTriggers:
            if trigger.matches(searchString):
                return True
        return False


class TextFilter(Filter, TextTriggerGroup):
    def __init__(self, filterText, *args):
        TextTriggerGroup.__init__(self, self.parseInput(filterText, *args))

    def parseInput(self, filterText, *args):
        return commasplit(filterText)

    def acceptsTestSuiteContents(self, suite):
        return not suite.isEmpty()


class ApplicationFilter(TextFilter):
    option = "app"

    def acceptsTestCase(self, test):
        return self.acceptsApplication(test.app)

    def acceptsTestSuite(self, suite):
        return self.acceptsApplication(suite.app)

    def acceptsApplication(self, app):
        return self.stringContainsText(app.name)


class TestSelectionFilter(TextFilter):
    option = "tp"

    def __init__(self, *args):
        self.diag = logging.getLogger("TestSelectionFilter")
        self.fullSuites = []
        TextFilter.__init__(self, *args)

    def parseInput(self, filterText, app, suites):
        allEntries = TextFilter.parseInput(self, filterText, app, suites)
        if os.name == "nt":
            allEntries = [ entry.replace("/", os.sep) for entry in allEntries ] # read saved filters from UNIX on Windows
        self.diag.info("allEntries " + repr(allEntries))
        if allEntries[0].startswith("appdata="):
            # chopped up per application
            return self.parseForApp(allEntries, app, suites)
        else:
            # old style, one size fits all
            return allEntries

    def parseForApp(self, allEntries, app, suites):
        active = False
        myEntries = []
        toFind = self.getSectionsToFind(allEntries, app, suites)
        self.diag.info("Sections for " + repr(app) + " = " + repr(toFind))
        for entry in allEntries:
            if entry in toFind:
                active = True
            elif entry.startswith("appdata="):
                active = False
            elif active:
                myEntries.append(entry)
        self.diag.info("Found " + repr(myEntries) + " from " + repr(allEntries))
        return myEntries

    def getSectionsToFind(self, allEntries, app, suites):
        allHeaders = [entry for entry in allEntries if entry.startswith("appdata=" + app.name)]
        if len(allHeaders) <= 1:
            return allHeaders
        allApps = [a for a in [suite.app for suite in suites] if a.name == app.name]
        sections = []
        for header in allHeaders:
            versionSet = set(header.split(".")[1:])
            self.diag.info("Looking for app matching " + repr(versionSet))
            bestApp = self.findAppMatchingVersions(versionSet, allApps)
            self.diag.info("Best app for " + header + " = " + repr(bestApp))
            if bestApp is app:
                sections.append(header)

        if len(sections) == 0:
            # We aren't a best-fit for any of them, so we do our best to find one anyway...
            return self.findSectionsMatchingApp(app, allHeaders)
        elif len(sections) == 1:
            return sections

        return self.findSectionsMatchingApp(app, sections)

    def findSectionsMatchingApp(self, app, allHeaders):
        myVersionSet = set(app.versions)

        def matchKey(header):
            currVersionSet = set(header.split(".")[1:])
            return self.getVersionSetMatchKey(currVersionSet, myVersionSet)

        # Really want some kind of "multimax" here
        bestHeader = max(allHeaders, key=matchKey)
        bestKey = matchKey(bestHeader)
        return [h for h in allHeaders if matchKey(h) == bestKey]

    def getVersionSetMatchKey(self, vset1, vset2):
        return len(vset1.intersection(vset2)), -len(vset1.symmetric_difference(vset2))

    def findAppMatchingVersions(self, myVersionSet, allApps):
        def matchKey(app):
            return self.getVersionSetMatchKey(set(app.versions), myVersionSet)
        return max(allApps, key=matchKey)

    def acceptsTestCase(self, test):
        return test.getRelPath() in self.texts or self.hasFullSuiteAncestor(test.parent)

    def hasFullSuiteAncestor(self, suite):
        return suite in self.fullSuites or (suite.parent and self.hasFullSuiteAncestor(suite.parent))

    def acceptsTestSuite(self, suite):
        return self.suiteInTexts(suite) or (suite.parent and self.hasFullSuiteAncestor(suite.parent))

    def suiteInTexts(self, suite):
        if suite.parent is None:
            return True  # don't eliminate the root suite :)
        for relPath in self.texts:
            suitePath = suite.getRelPath()
            if relPath == suitePath:
                self.fullSuites.append(suite)
                return True
            elif relPath.startswith(suitePath + os.sep):
                return True
        return False


# Generic action to be performed: all actions need to provide these methods
class Action:
    def __call__(self, test):
        pass

    def setUpSuite(self, suite):
        pass

    def setUpApplication(self, app):
        pass

    def tearDownSuite(self, suite):
        pass

    def kill(self, test, sig):
        pass

    def callDuringAbandon(self, testArg):
        # set to True if tests should have this action called even after all is reckoned complete (e.g. UNRUNNABLE)
        return False
    # Useful for printing in a certain format...

    def describe(self, testObj, postText=""):
        log.info(testObj.getIndent() + repr(self) + " " + repr(testObj) + postText)

    def __str__(self):
        return str(self.__class__)

    @classmethod
    def finalise(cls):
        pass


class ScriptWithArgs(Action):
    @classmethod
    def parseArguments(cls, args, allowedArgs):
        currKey = ""
        dict = {}
        for arg in args:
            if "=" in arg:
                newKey, val = arg.split("=", 1)
                if newKey in allowedArgs:
                    currKey = newKey
                    dict[currKey] = val
                elif currKey in dict:
                    dict[currKey] += " " + arg
                else:
                    print("Unrecognised option '" + newKey + "'")
            elif currKey in dict:
                dict[currKey] += " " + arg
        return dict


def addCategory(name, briefDesc, longDesc=""):
    if longDesc:
        TestState.categoryDescriptions[name] = briefDesc, longDesc
    else:
        TestState.categoryDescriptions[name] = briefDesc, briefDesc

# Observer mechanism shouldn't allow for conflicting notifications. Use main thread at all times


class ThreadedNotificationHandler:
    def __init__(self):
        self.workQueue = Queue()
        self.mutex = RLock()
        self.active = False
        self.allowedEvents = []
        self.idleHandler = None
        self.source = None

    def blockEventsExcept(self, allowedEvents):
        self.allowedEvents = allowedEvents

    def enablePoll(self, idleHandleMethod, **kwargs):
        self.active = True
        self.idleHandler = lambda : idleHandleMethod(self.pollQueue, **kwargs)

    def disablePoll(self, idleHandleRemover):
        with self.mutex:
            if self.source is not None:
                idleHandleRemover(self.source)
                self.source = None
            self.idleHandler = None

    def pollQueue(self):
        with self.mutex:
            try:
                observable, args, kwargs = self.workQueue.get_nowait()
                if len(self.allowedEvents) == 0 or args[0] in self.allowedEvents:
                    observable.diagnoseObs("From work queue", *args, **kwargs)
                    observable.performNotify(*args, **kwargs)
            except Empty:
                self.source = None
                return False
            return True

    def transfer(self, observable, *args, **kwargs):
        with self.mutex:
            self.workQueue.put((observable, args, kwargs))
            if self.active and self.source is None and self.idleHandler is not None:
                self.source = self.idleHandler()


class Observable:
    threadedNotificationHandler = ThreadedNotificationHandler()
    obsDiag = None
    LAST_OBSERVER = "last observer"

    @classmethod
    def diagnoseObs(klass, message, *args, **kwargs):
        if not klass.obsDiag:
            klass.obsDiag = logging.getLogger("Observable")
        klass.obsDiag.info(message + " " + str(klass) + " " + repr(args) + repr(kwargs))

    def __init__(self, passSelf=False):
        self.observers = []
        self.passSelf = passSelf

    def addObserver(self, observer):
        self.observers.append(observer)

    def setObservers(self, observers):
        self.observers = [x for x in observers if x is not self]

    def inMainThread(self):
        return currentThread().getName() == "MainThread"

    def notify(self, *args, **kwargs):
        if self.threadedNotificationHandler.active and not self.inMainThread():
            self.diagnoseObs("To work queue", *args, **kwargs)
            self.threadedNotificationHandler.transfer(self, *args, **kwargs)
        else:
            self.diagnoseObs("Perform directly", *args, **kwargs)
            self.performNotify(*args, **kwargs)

    def notifyThreaded(self, *args, **kwargs):
        # join the idle handler queue even if we're the main thread
        if self.threadedNotificationHandler.active:
            self.diagnoseObs("To work queue explicitly", *args, **kwargs)
            self.threadedNotificationHandler.transfer(self, *args, **kwargs)
        else:
            self.diagnoseObs("Perform directly", *args, **kwargs)
            self.performNotify(*args, **kwargs)

    def notifyIfMainThread(self, *args, **kwargs):
        if not self.inMainThread():
            return
        else:
            self.diagnoseObs("Perform directly", *args, **kwargs)
            self.performNotify(*args, **kwargs)

    def performNotify(self, name, *args, **kwargs):
        methodName = "notify" + name
        lastObserver = None
        for observer in self.observers:
            if hasattr(observer, methodName):
                self.diagnoseObs("Notify observer " + name + " " + str(observer.__class__))
                answer = self.notifyObserver(observer, methodName, *args, **kwargs)
                if answer == self.LAST_OBSERVER:
                    self.diagnoseObs("Setting as last observer", *args, **kwargs)
                    lastObserver = observer
        if lastObserver:
            self.diagnoseObs("Notify last observer", *args, **kwargs)
            lastObserver.notifyLastObserver(methodName)

    def notifyObserver(self, observer, methodName, *args, **kwargs):
        # doesn't matter if only some of the observers have the method
        method = getattr(observer, methodName)
        # unpickled objects have not called __init__, and
        # hence do not have self.passSelf ...
        try:
            if hasattr(self, "passSelf") and self.passSelf:
                return method(self, *args, **kwargs)
            else:
                return method(*args, **kwargs)
        except TextTestException:
            raise
        except Exception:
            sys.stderr.write("Observer raised exception while calling " + methodName + " :\n")
            printException()

# Interface all responders must fulfil


class Responder:
    def __init__(self, *args):
        pass

    def addSuites(self, suites):
        for suite in suites:
            self.addSuite(suite)
    # Full suite of tests, get notified of it at the start...

    def addSuite(self, suite):
        pass
    # Called when the state of the test "moves on" in its lifecycle

    def notifyLifecycleChange(self, test, state, changeDesc):
        pass
    # Called when no further actions will be performed on the test

    def notifyComplete(self, test):
        pass
    # Called when everything is finished

    def notifyAllComplete(self):
        pass

    def canBeMainThread(self):
        return True


# Generic state which tests can be in, should be overridden by subclasses
# Acts as a static state for tests which have not run (yet)
# Free text is text of arbitrary length: it will appear in the "Text Info" GUI window when the test is viewed
# and the "details" section in batch mode
# Brief text should be at most two or three words: it appears in the details column in the main GUI window and in
# the summary of batch mode
class TestState(Observable):
    categoryDescriptions = OrderedDict()
    showExecHosts = 0
    defaultBriefText = ""

    def __init__(self, category, freeText="", briefText="", started=0, completed=0,
                 executionHosts=[], lifecycleChange=""):
        Observable.__init__(self)
        self.category = category
        self.freeText = freeText
        self.briefText = briefText
        self.started = started
        self.completed = completed
        self.executionHosts = executionHosts
        self.lifecycleChange = lifecycleChange

    def __str__(self):
        return self.freeText

    def __repr__(self):
        return self.categoryRepr() + self.hostRepr() + self.colonRepr()

    def categoryRepr(self):
        if self.category not in self.categoryDescriptions:
            return self.category
        longDescription = self.categoryDescriptions[self.category][1]
        return longDescription

    def hostString(self):
        return "on " + ", ".join(self.executionHosts)

    def hostRepr(self):
        if self.showExecHosts and len(self.executionHosts) > 0:
            return " " + self.hostString()
        else:
            return ""

    def colonRepr(self):
        if self.hasSucceeded():
            return ""
        else:
            return " :"

    def getComparisonsForRecalculation(self):
        # Is some aspect of the state out of date
        return []

    # Used by text interface to print states
    def description(self):
        if "\n" in self.freeText:
            return "not compared:\n" + self.freeText
        else:
            return "not compared:  " + self.freeText

    def getFreeText(self):
        return self.freeText  # some subclasses might want to calculate this...

    def getTypeBreakdown(self):
        if self.isComplete():
            return "failure", self.briefText
        else:
            return self.category, self.briefText

    def hasStarted(self):
        return self.started or self.completed

    def isComplete(self):
        return self.completed

    def hasSucceeded(self):
        return 0

    def hasFailed(self):
        return self.isComplete() and not self.hasSucceeded()

    def getExitCode(self):
        return int(self.hasFailed())

    def isMarked(self):
        return False

    def hasResults(self):
        # Do we have actual results that can be compared
        return 0

    def shouldAbandon(self):
        return self.lifecycleChange == "complete"

    def isSaveable(self):
        return self.hasFailed() and self.hasResults()

    def warnOnSave(self):  # pragma : no cover - only called on saveable tests usually
        return False

    def updateAfterLoad(self, app, **kwargs):
        pass

    def makeModifiedState(self, *args):
        pass


addCategory("unrunnable", "unrunnable", "could not be run")
addCategory("marked", "marked", "was marked by the user")


class Unrunnable(TestState):
    def __init__(self, freeText, briefText, executionHosts=[], lifecycleChange=""):
        TestState.__init__(self, "unrunnable", freeText, briefText, completed=1,
                           executionHosts=executionHosts, lifecycleChange=lifecycleChange)

    def shouldAbandon(self):
        return True


class MarkedTestState(TestState):
    def __init__(self, freeText, briefText, oldState, executionHosts=[]):
        self.oldState = oldState
        self.myFreeText = freeText
        fullText = freeText + "\n\nORIGINAL STATE:\nTest " + repr(oldState) + "\n " + oldState.freeText
        TestState.__init__(self, "marked", fullText, briefText, completed=1,
                           executionHosts=executionHosts, lifecycleChange="marked")

    # We must implement this ourselves, since we want to be neither successful nor
    # failed, and by default hasFailed is implemented as 'not hasSucceeded()'.
    def hasFailed(self):
        return False

    def hasResults(self):
        return True

    def isMarked(self):
        return True

    def getTypeBreakdown(self):
        return self.category, self.briefText

    def getComparisonsForRecalculation(self):
        # Is some aspect of the state out of date
        return self.oldState.getComparisonsForRecalculation()

    def __getattr__(self, name):
        # Anything not implemented should be called on the actual state...
        return getattr(self.oldState, name)

    def makeNewState(self, *args):
        newOldState = self.oldState.makeNewState(*args)
        return MarkedTestState(self.myFreeText, self.briefText, newOldState, self.executionHosts)


class TestStateUnpickler(Unpickler):
    def find_class(self, modName, className):
        try:
            namespace = {}
            exec("from " + modName + " import " + className + " as _class", globals(), namespace)
            return namespace["_class"]
        except ImportError as e:
            if not modName.startswith("texttestlib"):
                try:
                    return self.find_class("texttestlib." + modName, className)
                except:
                    raise e
            else:
                raise e


def getNewTestStateFromFile(file):
    unpickler = TestStateUnpickler(file)
    try:
        return unpickler.load()
    except Exception:
        encoding = getpreferredencoding()
        from io import BytesIO
        file.seek(0)
        unpickler = TestStateUnpickler(BytesIO(file.read().replace(b"\r\n", b"\n")), encoding=encoding, errors="replace")
        return unpickler.load()
    

log = None


def configureLogging(configFile=None):
    # only set up once
    global log
    if not log:
        if configFile:
            # First is for TextTest troubleshooting
            # Second is for self-tests
            # There appears to be a bug on Windows here, along the lines of https://bugs.python.org/issue19528. All backslashes double up!
            defaults = {"TEXTTEST_PERSONAL_LOG": getPersonalDir("log").replace("\\", "\\\\"),
                        "TEXTTEST_LOG_DIR": os.getenv("TEXTTEST_LOG_DIR", "").replace("\\", "\\\\")}
            logging.config.fileConfig(configFile.replace("\\", "\\\\"), defaults)
        log = logging.getLogger("standard log")


def getPersonalConfigDir():
    return os.getenv("TEXTTEST_PERSONAL_CONFIG")

# Return the hostname, guaranteed to be just the hostname...


def gethostname():
    fullname = socket.gethostname()
    return fullname.split(".")[0]

# Return 'localhost' if it is the local host...


def interpretHostname(hostname):
    if hostname == "localhost" or len(hostname) == 0:
        return "localhost"
    localhost = gethostname()
    if hostsMatch(hostname, localhost):
        return "localhost"
    else:
        return hostname


def hostsMatch(hostname, localhost):
    if "@" in hostname:
        user, host = hostname.split("@")
        return hostsMatch(host, localhost) and user == os.getenv("USER")
    else:
        parts = hostname.split(".")
        return parts[0] == localhost


# Hacking around os.path.getcwd not working with AMD automounter
def abspath(path):
    if "PWD" in os.environ:
        return os.path.join(os.environ["PWD"], path)
    else:
        return os.path.abspath(path)

# deepcopy(os.environ) still seems to retain links to the actual environment, create a clean copy


def copyEnvironment(values={}, ignoreVars=[]):
    for var, value in list(os.environ.items()):
        if var not in values and var not in ignoreVars:
            values[var] = value
    return values


def getInterpreter(executable):
    extension = executable.rsplit(".", 1)[-1]
    cache = {"py": "python",
             "rb": "ruby",
             "jar": "java -jar"}
    return cache.get(extension, "")


def commandLineString(cmdArgs, defaultQuoteChar='"'):
    def getQuoteChar(char):
        if char == "\"" and os.name == "posix":
            return "'"
        else:
            return defaultQuoteChar

    def quoteArg(arg):
        if len(arg) == 0:
            return '""'
        quoteChars = "'\"|* "
        for char in quoteChars:
            if char in arg:
                quoteChar = getQuoteChar(char)
                return quoteChar + arg + quoteChar
        return arg

    return " ".join(map(quoteArg, cmdArgs))


def relpath(fullpath, parentdir, normalise=True):
    normFull = os.path.normpath(fullpath) if normalise else fullpath
    normParent = os.path.normpath(parentdir) if normalise else parentdir
    relPath = normFull.replace(normParent, "")
    if relPath != normFull:
        if relPath.startswith(os.sep):
            return relPath[1:]
        else:
            return relPath


def copyPath(srcPath, dstPath):
    if os.path.isdir(srcPath):
        removePath(dstPath)
        shutil.copytree(srcPath, dstPath)
    else:
        shutil.copy(srcPath, dstPath)


def removePath(path):
    if os.path.isfile(path) or os.path.islink(path):
        os.remove(path)
        return True
    elif os.path.isdir(path):
        shutil.rmtree(path)
        return True
    else:
        return False

# Useful utility, free text input as comma-separated list which may have spaces


def commasplit(input):
    return list(map(str.strip, input.split(",")))

# Another useful thing that saves an import and remembering weird stuff


def modifiedTime(filename):
    try:
        return os.stat(filename)[stat.ST_MTIME]
    except OSError:
        # Dead links etc.
        return None

# portable version of shlex.split
# See http://sourceforge.net/tracker/index.php?func=detail&aid=1724366&group_id=5470&atid=105470
# As usual, UNIX paths work better on Windows than Windows paths...


def splitcmd(s):
    if os.name == "posix":
        return shlex.split(s)
    else:
        if "Program Files" in s: # such a common case on Windows, don't mess up
            return [ s ]
        else:
            return shlex.split(s.replace("\\", "\\\\"))

# portable version of os.path.samefile


def samefile(writeDir, currDir):
    try:
        return os.path.samefile(writeDir, currDir)
    except Exception:  # AttributeError for Windows, OSError if currDir doesn't exist
        # samefile doesn't exist on Windows, but nor do soft links so we can
        # do a simpler version
        return os.path.normpath(writeDir.replace("\\", "/")) == os.path.normpath(currDir.replace("\\", "/"))


def makeWriteable(path):
    currMode = os.stat(path)[stat.ST_MODE]
    currPerm = stat.S_IMODE(currMode)
    newPerm = currPerm | 0o220
    os.chmod(path, newPerm)


def getPaths(d):
    paths = [d]
    for root, dirs, files in os.walk(d):
        for path in dirs + files:
            fullPath = os.path.join(root, path)
            if not os.path.islink(fullPath):
                # Don't want to / can't change permissions of links anyway
                paths.append(fullPath)
    return paths

# Version of rmtree not prone to crashing if directory in use or externally removed


def rmtree(dir, attempts=100):
    realDir = os.path.realpath(dir)
    if not os.path.isdir(realDir):
        log.info("Write directory " + dir + " externally removed")
        return
    # Don't be somewhere under the directory when it's removed
    try:
        if os.getcwd().startswith(realDir):
            root = os.path.dirname(os.path.normpath(dir))
            os.chdir(root)
    except OSError:  # pragma: no cover - robustness only
        pass
    for i in range(attempts):
        try:
            shutil.rmtree(realDir)
            return
        except PermissionError as e:
            # We own this stuff, don't respect readonly flags set by ourselves, it might just be the SUT doing so...
            for path in getPaths(realDir):
                try:
                    makeWriteable(path)
                except OSError as e:
                    log.info("Could not change permissions to be able to remove directory " +
                             dir + " : - " + str(e))
                    return
        except Exception as e:
            if os.path.isdir(realDir):
                if i == attempts - 1:
                    log.info("Unable to remove directory " + dir + " :")
                    printException()
                else:
                    log.info("Problems removing directory " + dir + " - waiting 1 second to retry...")
                    time.sleep(1)


def fakeMissingFileText():
    return b"Auto-generated by TextTest to simulate missing file for this version...\n"


def containsAutoGeneratedText(resultFile):
    firstLine = open(resultFile, "rb").readline()
    return firstLine.strip() == fakeMissingFileText().strip()


class AggregationError(RuntimeError):
    def __init__(self, value1, value2, index):
        self.value1 = value1
        self.value2 = value2
        self.index = index


# Useful utility for combining different response values for a series of method calls
class ResponseAggregator:
    def __init__(self, methods):
        self.methods = methods

    def __call__(self, *args, **kwargs):
        if len(self.methods) == 0:
            return

        basicValue = self.methods[0](*args, **kwargs)
        for i, extraMethod in enumerate(self.methods[1:]):
            extraValue = extraMethod(*args, **kwargs)
            if type(basicValue) == list:
                for item in extraValue:
                    if not item in basicValue:
                        basicValue.append(item)
            elif type(basicValue) == dict:
                basicValue.update(extraValue)
            elif extraValue != basicValue:
                if self.isSet(extraValue):
                    if self.isSet(basicValue):
                        raise AggregationError(basicValue, extraValue, i + 1)
                    else:
                        basicValue = extraValue

        return basicValue

    def isSet(self, value):
        return value != "" and value != "disabled"


def getAggregateString(items, method):
    values = []
    for item in items:
        value = method(item)
        if value is not None and value not in values:
            values.append(value)

    if len(values) > 1:
        return "<default> - " + ",".join(values)
    else:
        return values[0]


def readList(filename):
    try:
        items = []
        for longline in open(filename).readlines():
            line = longline.strip()
            if len(line) > 0 and not line.startswith("#"):
                items.append(line)
        return items
    except IOError:
        return []  # It could be a broken link: don't bail out if so...


def readListWithComments(filename, filterMethod=None):
    items = OrderedDict()
    badItems = OrderedDict()
    currComment = ""
    emptyLineSymbol = "__EMPTYLINE__"

    for longline in open(filename, encoding=getpreferredencoding(), errors="replace").readlines():
        line = longline.strip()
        if len(line) == 0:
            if currComment:
                currComment = currComment.strip() + emptyLineSymbol
            continue
        if line.startswith("#"):
            if emptyLineSymbol in currComment:
                items[currComment] = currComment.replace(emptyLineSymbol, "")
                currComment = ""
            currComment += longline.lstrip()
        else:
            if filterMethod:
                failReason, warningText = filterMethod(line, items, filename)
                if failReason:
                    currComment += "# " + line + " (automatically commented due to " + failReason + ")"
                    badItems[line] = warningText
                    items["#" + line] = currComment
                    currComment = ""
                    continue
            if emptyLineSymbol in currComment:
                items[currComment] = currComment.replace(emptyLineSymbol, "")
                currComment = ""
            items[line] = uncomment(currComment) if currComment else ""
            currComment = ""
    # Rescue dangling comments in the end
    if currComment and len(items) > 0:
        items[currComment] = currComment.replace(emptyLineSymbol, "")
    return items, badItems


def uncomment(comment):
    lines = comment.split("\n")
    return "\n".join([line[1:].lstrip() for line in lines if line.startswith("#")])


def openForWrite(path, mode="w"):
    ensureDirExistsForFile(path)
    return open(path, mode)

# Make sure the dir exists


def ensureDirExistsForFile(path):
    dir = os.path.split(path)[0]
    ensureDirectoryExists(dir)


def addLocalPrefix(fullPath, prefix):
    dir, file = os.path.split(fullPath)
    return os.path.join(dir, prefix + "_" + file)


def ensureDirectoryExists(path, attempts=5):
    # os.makedirs seems to be a bit flaky, especially if the file server is loaded
    # or on silly platforms like powerpc. We give it five chances to do its stuff :)
    for attempt in range(attempts):
        if os.path.isdir(path):
            return
        try:
            os.makedirs(path)
        except OSError:
            if attempt == attempts - 1:
                raise


def retryOnInterrupt(function, *args):
    try:
        return function(*args)
    except (IOError, OSError) as detail:
        if str(detail).find("Interrupted system call") != -1:
            return retryOnInterrupt(function, *args)
        else:
            raise


def tryFileChange(function, permissionMessage, *args):
    try:
        return function(*args)
    except OSError as e:
        errorStr = str(e)
        if "Permission" in errorStr:
            raise TextTestError(permissionMessage)
        else:
            raise TextTestError(errorStr)


def getExceptionString():
    type, value, traceback = sys.exc_info()
    return "".join(format_exception(type, value, traceback))


def printException():
    sys.stderr.write("Description of exception thrown :\n")
    exceptionString = getExceptionString()
    sys.stderr.write(exceptionString)
    sys.stderr.flush()
    return exceptionString


def zeroDivisorPercentage(numerator):
    if numerator == 0.0:
        return 0
    else:
        return -1


def calculatePercentageNormalised(oldVal, newVal):
    largest = max(oldVal, newVal)
    smallest = min(oldVal, newVal)
    if smallest != 0.0:
        return ((largest - smallest) / abs(smallest)) * 100
    else:
        return zeroDivisorPercentage(largest)


def calculatePercentageStandard(oldVal, newVal):
    if oldVal != 0.0:
        diff = abs(newVal - oldVal)
        return (diff / abs(oldVal)) * 100
    else:
        return zeroDivisorPercentage(newVal)


def roundPercentage(val):
    perc = int(round(val))
    if perc == 0:
        return float("%.0e" % val)  # Print one significant figure
    else:
        return perc


class PreviewGenerator:
    def __init__(self, maxWidth, maxLength):
        self.maxWidth = maxWidth
        self.maxLength = maxLength

    def getCutLines(self, lines):
        if len(lines) < self.maxLength:
            return lines
        else:
            return lines[:self.maxLength] + ["<truncated after showing first " + pluralise(self.maxLength, "line") + ">\n"]

    def getPreview(self, file):
        fileLines = retryOnInterrupt(self.getFileLines, file)
        return self.getPreviewFromLines(fileLines)

    def getFileLines(self, file):
        lines = file.readlines()
        file.close()
        return lines

    def getPreviewFromLines(self, lines):
        cutLines = self.getCutLines(lines)
        lines = list(map(self.getWrappedLine, cutLines))
        return "".join(lines)

    def getWrappedLine(self, line):
        remaining = line
        result = ""
        while True:
            if len(remaining) <= self.maxWidth:
                return result + remaining
            result += remaining[:self.maxWidth] + "\n"
            remaining = remaining[self.maxWidth:]


class TextTestException(RuntimeError):
    pass

# Exception to throw. It's generally good to throw this internally


class TextTestError(TextTestException):
    pass


class TextTestWarning(TextTestException):
    pass

# Sort of a workaround to get e.g. CVSLogInGUI to show a message in a simple info dialog


class TextTestInformation(RuntimeError):
    pass

# Yes, we know that getopt exists. However it throws exceptions when it finds unrecognised things, and we can't do that...


class OptionFinder(OrderedDict):
    def __init__(self, args, defaultKey="default"):
        OrderedDict.__init__(self)
        self.buildOptions(args, defaultKey)

    def buildOptions(self, args, defaultKey):
        optionKey = None
        for item in args:
            if item.startswith("-"):
                optionKey = self.stripMinuses(item)
                self[optionKey] = None
            elif optionKey:
                if self[optionKey] is not None:
                    self[optionKey] += " "
                else:
                    self[optionKey] = ""
                self[optionKey] += item.strip()
            else:
                self[defaultKey] = item

    def stripMinuses(self, item):
        if len(item) > 1 and item[1] == "-":
            return item[2:].strip()
        else:
            return item[1:].strip()


class TextTrigger:
    def __init__(self, text, tryAsRegexp=True, matchEmptyString=True):
        self.text = text
        self.regex = None
        self.matchEmptyString = matchEmptyString
        if tryAsRegexp and isRegularExpression(text):
            try:
                self.regex = re.compile(text)
            except re.error:
                pass

    def __repr__(self):
        return self.text

    def matches(self, line, *args):
        if self.regex:
            return self.regex.search(line)
        else:
            found = line.find(self.text) != -1
            if not self.matchEmptyString:
                found = (found and self.text != "") or line == self.text
            return found

    def replace(self, line, newText):
        if self.regex:
            return re.sub(self.text, newText, line)
        else:
            return line.replace(self.text, newText)

    def reset(self):
        pass

class PatternAggregator:
    def __init__(self):
        self.groups = 0
        self.groupindex = {}

    def add(self, pattern):
        self.groups += pattern.groups
        self.groupindex.update(pattern.groupindex)


class MatchAggregator:
    def __init__(self):
        self.match = None
        self.groups = ()
        self.patternAggregator = PatternAggregator()

    def reset(self):
        self.match = None
        self.groups = ()

    def group(self, i):
        return self.groups[i - 1]

    def add(self, match):
        if hasattr(match, "group"):
            self.match = match
            self.groups += match.groups()
            self.patternAggregator.add(match.re)

    def expand(self, template):
        if self.match is not None:
            # ?? This is not documented, but has remained unchanged across many Python versions
            # Better ideas welcome. The match objects are defined in C, the class name isn't public and hence they cannot be subclassed or monkey patched
            return re._expand(self.patternAggregator, self, template)
        else:
            return template

    def __getattr__(self, name):
        return getattr(self.match, name)


class MultilineTextTrigger(TextTrigger):
    def __init__(self, text, tryAsRegexp, matchEmptyString=True):
        TextTrigger.__init__(self, text, False, matchEmptyString)
        self.triggers = []
        self.currentIndex = 0
        self.matchedLines = []
        self.aggregator = MatchAggregator()
        self.trailingNewline = text.endswith("\n")
        lines = text.rstrip("\n").split("\n") if text else []
        for line in lines:
            self.triggers.append(TextTrigger(line, tryAsRegexp, matchEmptyString))

    def matches(self, line):
        return self._matches(line)[0]

    def _matches(self, line):
        match = self.triggers[self.currentIndex].matches(line)
        if match:
            self.aggregator.add(match)
            self.currentIndex += 1
            if self.currentIndex == len(self.triggers):
                self.currentIndex = 0
                return True, True
            else:
                return False, True
        return False, False

    def replace(self, line, newTextLines):
        matchComplete, matchThisLine = self._matches(line)
        if matchThisLine:
            self.matchedLines.append(line)
            text = ""
            if matchComplete:
                for i in range(len(newTextLines)):
                    if i < len(self.matchedLines):
                        newText = self.aggregator.expand(newTextLines[i])
                        newLine = self.triggers[i].replace(self.matchedLines[i], newText)
                        if self.trailingNewline and i == len(newTextLines) - 1:
                            newLine = newLine.rstrip("\n")
                        text += newLine
                    else:
                        text += newTextLines[i] + "\n"
                self.reset()
            return text
        else:
            if not self.matchEmptyString:
                # We want to remove matched lines if the current trigger text is the empty string.
                text = "".join(self.matchedLines) + line if self.text[-1] != "\n" else line
            self.reset()
            return text

    def getLeftoverText(self):
        text = "".join(self.matchedLines)
        self.reset()
        return text

    def reset(self):
        self.currentIndex = 0
        self.matchedLines = []
        self.aggregator.reset()


# Used for application and personal configuration files
class MultiEntryDictionary(OrderedDict):
    warnings = []

    def __init__(self, importKey="", importFileFinder=None, aliases={}, allowSectionHeaders=True, fileTrackSections={}, *args, **kw):
        OrderedDict.__init__(self, *args, **kw)
        self.diag = logging.getLogger("MultiEntryDictionary")
        self.aliases = aliases
        self.importKey = importKey
        self.importFileFinder = importFileFinder
        self.allowSectionHeaders = allowSectionHeaders
        self.fileTrackSections = fileTrackSections

    def __reduce__(self):
        # Need this because of __reduce__ in OrderedDict
        # which merrily reduces all the attributes of the subclass without asking
        # Used when doing deepcopy()
        items = [[k, self[k]] for k in self]
        return self.__class__, (self.importKey, Callable(self.importFileFinder),
                                self.aliases, self.allowSectionHeaders, self.fileTrackSections, items)

    def addFileTracking(self, key):
        self.fileTrackSections[key] = {}

    def getSectionInfo(self, sectionName=""):
        if sectionName and sectionName != "end":
            return self[sectionName], sectionName
        else:
            return self, "<global>"

    def setAlias(self, aliasName, realName):
        self.aliases[aliasName] = realName

    def getEntryName(self, fromConfig):
        return self.aliases.get(fromConfig, fromConfig)

    def readValues(self, fileNames, *args, **kwargs):
        for filename in fileNames:
            self.readFromFile(filename, *args, **kwargs)

    def readFromFile(self, filename, *args, **kwargs):
        self.diag.info("Reading file " + filename)
        currSectionName = ""
        for line in readList(filename):
            if self.allowSectionHeaders and self.isSectionHeader(line):
                currSectionName = self.getNewSectionInfo(line, *args, **kwargs)
            elif ":" in line:
                self.parseConfigLine(line, currSectionName, filename, *args, **kwargs)
            else:
                self.warn("Could not parse config line " + line)

    def isSectionHeader(self, line):
        return line.startswith("[") and line.endswith("]")

    def warn(self, message):
        if message not in self.warnings:
            self.warnings.append(message)
            printWarning(message)

    def getFileDefining(self, sectionName, entryName, value, envMapping=os.environ):
        valueDict = self.fileTrackSections.get(sectionName)
        if valueDict:
            valueToUse = self.findFileDefiningValue(value, valueDict, envMapping)
            for currEntry, filename in valueDict.get(valueToUse, []):
                if fnmatch.fnmatch(entryName, currEntry):
                    return filename, currEntry
        return None, None

    def findFileDefiningValue(self, value, valueDict, envMapping):
        if value in valueDict:
            return value
        for storedValue in list(valueDict.keys()):
            if "$" in storedValue and self.expandEnvironment(storedValue, envMapping) == value:
                return storedValue

    def parseConfigLine(self, line, currSectionName, filename, *args, **kwargs):
        key, value = line.split(":", 1)
        entryName = self.getEntryName(string.Template(key).safe_substitute(os.environ))
        self.addEntry(entryName, value, currSectionName, *args, **kwargs)
        if currSectionName in self.fileTrackSections:
            self.fileTrackSections[currSectionName].setdefault(value, []).append((entryName, filename))
        if key and key == self.importKey:
            self.readFromFile(self.importFileFinder(os.path.expandvars(value)), *args, **kwargs)

    def getNewSectionInfo(self, line, insert=True, errorOnUnknown=False):
        name = self.getEntryName(line[1:-1])
        if name != "end":
            if name in self:
                value = self[name]
                if isinstance(value, OrderedDict):
                    return name
                elif type(value) == dict:
                    # important, e.g. for run_dependent_text, replace all dictionaries with ordered versions
                    self[name] = OrderedDict(value)
                    return name
                else:
                    self.warn("Config entry name '" + name + "' incorrectly used as a section marker.")
            elif insert:
                self[name] = OrderedDict()
                return name
            elif errorOnUnknown:
                self.warn("Config section name '" + name + "' not recognised.")
        return ""

    def addEntry(self, entryName, entry, sectionName="", *args, **kwargs):
        currDict, currSection = self.getSectionInfo(sectionName)
        try:
            self._addEntry(entryName, entry, currDict, currSection, *args, **kwargs)
        except ValueError:
            self.warn("Config entry name '" + entryName + "' in section '" + currSection +
                      "' given an invalid value '" + entry + "', ignoring.")

    def removeEntry(self, entryName, entry, sectionName=""):
        currDict, _ = self.getSectionInfo(sectionName)
        if entryName in currDict:
            dictElem = currDict[entryName]
            if entry in dictElem:
                dictElem.remove(entry)

    def _addEntry(self, entryName, entry, currDict, currSection,
                  insert=True, errorOnUnknown=False, errorOnClashWithGlobal=True):
        if currDict is not self and entryName in self and errorOnClashWithGlobal:
            self.warn("Config entry name '" + entryName + "' found in section '" + currSection +
                      "', but defined at global scope. Did you forget an [end] marker?")

        entryExists = entryName in currDict
        if entryExists:
            self.diag.info("Entry existed, setting " + entryName + "=" + str(entry))
            self.insertEntry(entryName, entry, currDict)
        else:
            if insert or not currDict is self:
                self.diag.info("Inserting " + entryName + "=" + repr(entry))
                currDict[entryName] = self.castEntry(entryName, entry, currDict)
            elif errorOnUnknown:
                self.warn("Config entry name '" + entryName + "' not recognised.")

    def getDictionaryValueType(self, currDict):
        val = list(currDict.values())
        if len(val) == 0:
            return str
        else:
            return type(val[0])

    def castEntry(self, dummy, entry, currDict):
        if type(entry) not in (str, bytes):
            return entry
        dictValType = self.getDictionaryValueType(currDict)
        if dictValType == list:
            return self.getBasicList(entry)
        else:
            return dictValType(entry)

    def getBasicList(self, entry):
        if entry.startswith("{CLEAR"):
            return []
        else:
            return [entry]

    def getListValue(self, entry, currentList):
        if entry == "{CLEAR LIST}":
            return []
        elif entry.startswith("{CLEAR"):
            itemToRemove = entry[7:-1]
            if itemToRemove in currentList:
                currentList.remove(itemToRemove)
        elif entry not in currentList:
            self.diag.info("Get list value for " + entry + repr(currentList))
            currentList.append(entry)
        return currentList

    def insertEntry(self, entryName, entry, currDict):
        currEntry = currDict[entryName]
        if isinstance(currEntry, list):
            currDict[entryName] = self.getListValue(entry, currDict[entryName])
        elif isinstance(currEntry, dict):
            newCurrDict = self.getSectionInfo(entryName)[0]
            if "default" in newCurrDict:
                self.insertEntry("default", entry, newCurrDict)
            else:
                self.warn("Config entry name '" + entryName + "' is only valid as a section marker.")
        else:
            currDict[entryName] = type(currEntry)(entry)

    def getSingle(self, key, expandVars=True, envMapping=os.environ):
        value = self.get(key)
        if expandVars:
            return self.expandEnvironment(value, envMapping)
        else:
            return value

    def getComposite(self, key, subKey, expandVars=True, envMapping=os.environ, defaultKey="default"):
        value = self.getCompositeUnexpanded(key, subKey, defaultKey)
        if expandVars:
            return self.expandEnvironment(value, envMapping)
        else:
            return value

    def getCompositeUnexpanded(self, key, subKey, defaultSubKey="default"):
        dict = self.get(key)
        # If it wasn't a dictionary, return None
        if not hasattr(dict, "items"):
            return None
        listVal = []
        usingList = False
        for currSubKey, currValue in dict.items():
            if fnmatch.fnmatch(subKey, currSubKey):
                if type(currValue) == list:
                    listVal += currValue
                    usingList = True
                else:
                    return currValue
        # A certain amount of duplication here - hard to see how to avoid it
        # without compromising performance though...
        if subKey != defaultSubKey:
            defValue = dict.get(defaultSubKey)
            if defValue is not None:
                if type(defValue) == list:
                    listVal += defValue
                    return listVal
                else:
                    return defValue
        if usingList:
            return listVal

    @classmethod
    def expandEnvironment(cls, value, envMapping):
        if isinstance(value, str):
            return string.Template(value).safe_substitute(envMapping)
        elif isinstance(value, list):
            return [string.Template(element).safe_substitute(envMapping) for element in value]
        elif isinstance(value, dict):
            newDict = value.__class__()
            for key, val in list(value.items()):
                newDict[key] = cls.expandEnvironment(val, envMapping)
            return newDict
        else:
            return value


class Option:
    def __init__(self, name, value, description, changeMethod):
        self.name = name
        self.defaultValue = value
        self.valueMethod = None
        self.updateMethod = None
        self.description = description
        self.changeMethod = changeMethod

    def getValue(self):
        if self.valueMethod:
            try:
                return self.valueMethod()
            except ValueError:
                self.valueMethod = None

        return self.defaultValue

    def resetMethods(self):
        self.valueMethod = None
        self.updateMethod = None

    def resetDefault(self):
        if self.valueMethod:
            value = self.valueMethod()
            if value is not None:
                self.defaultValue = value
            self.resetMethods()

    def getCmdLineValue(self):
        return self.getValue()

    def setValue(self, value):
        self.defaultValue = value
        if self.updateMethod:
            self.updateMethod(value)

    def setMethods(self, valueMethod, updateMethod):
        self.valueMethod = valueMethod
        self.updateMethod = updateMethod

    def reset(self):
        if self.updateMethod:
            self.updateMethod(self.defaultValue)
        else:
            self.valueMethod = None

    def addPossibleValue(self, value):
        pass


class TextOption(Option):
    def __init__(self, name, value="", possibleValues=[], allocateNofValues=-1,
                 selectDir=False, selectFile=False, saveFile=False, possibleDirs=[],
                 description="", changeMethod=None, multilineEntry=False, minimum=0,
                 maximum=1000):
        Option.__init__(self, name, value, description, changeMethod)
        self.possValAppendMethod = None
        self.possValListMethod = None
        self.possibleValues = []
        self.nofValues = allocateNofValues
        self.selectDir = selectDir
        self.selectFile = selectFile
        self.saveFile = saveFile
        self.possibleDirs = possibleDirs
        self.clearMethod = None
        self.multilineEntry = multilineEntry
        self.minimum = minimum
        self.maximum = maximum
        self.setPossibleValues(possibleValues)

    def resetMethods(self):
        Option.resetMethods(self)
        self.possValAppendMethod = None
        self.possValListMethod = None
        self.clearMethod = None

    def setPossibleValuesMethods(self, appendMethod, getMethod):
        self.possValListMethod = getMethod
        if appendMethod:
            self.possValAppendMethod = appendMethod
            self.updatePossibleValues()

    def updatePossibleValues(self):
        if self.possValAppendMethod:
            for value in self.possibleValues:
                self.possValAppendMethod(value)

    def listPossibleValues(self):
        if self.possValListMethod:
            return self.possValListMethod()
        else:
            return self.possibleValues

    def addPossibleValue(self, value):
        if value not in self.possibleValues:
            self.possibleValues.append(value)

    def setValue(self, value):
        Option.setValue(self, value)
        if self.usePossibleValues():
            self.setPossibleValues(self.possibleValues)

    def setPossibleValues(self, values):
        if self.selectFile or (self.defaultValue in values):
            self.possibleValues = values
        else:
            self.possibleValues = [self.defaultValue] + values
        self.clear()
        self.updatePossibleValues()

    def getPossibleDirs(self):
        if self.selectDir or self.selectFile:
            return self.possibleDirs + self.possibleValues
        else:
            return self.possibleDirs

    def usePossibleValues(self):
        return self.selectDir or self.nofValues > 1 or len(self.possibleValues) > 1

    def setClearMethod(self, clearMethod):
        self.clearMethod = clearMethod

    def clear(self):
        if self.clearMethod:
            self.clearMethod()

    def getDirectories(self):
        allDirs = self.getPossibleDirs()
        for dir in allDirs:
            try:
                ensureDirectoryExists(dir)
            except OSError:  # Might not have permissions
                allDirs.remove(dir)

        return allDirs, self.findDefaultDirectory(allDirs)

    def canBeDefaultDir(self, dir):
        return self.saveFile or len(os.listdir(dir)) > 0

    def getPreviousDirectory(self):
        prevVal = self.getValue()
        if prevVal and os.path.exists(prevVal):
            if self.selectDir:
                return prevVal
            else:
                return os.path.dirname(prevVal)

    def findDefaultDirectory(self, allDirs):
        previousDir = self.getPreviousDirectory()
        if previousDir in allDirs and self.canBeDefaultDir(previousDir):
            return previousDir

        # Set first non-empty dir as default ...)
        for dir in allDirs:
            if self.canBeDefaultDir(dir):
                return dir

        return allDirs[0]


class Switch(Option):
    def __init__(self, name="", value=0, options=[], hideOptions=False, description="", changeMethod=None, autoEnable=[]):
        Option.__init__(self, name, int(value), description, changeMethod)
        self.options = options
        self.hideOptions = hideOptions
        self.autoEnable = autoEnable

    def setValue(self, value):
        Option.setValue(self, int(value))

    def getValue(self):
        return int(Option.getValue(self))

    def toggle(self):
        self.setValue(1 - self.getValue())

    def getCmdLineValue(self):
        if len(self.options) > 2:
            return str(self.getValue())
        else:
            return ""  # always on or off...


class OptionGroup:
    def __init__(self, name):
        self.name = name
        self.options = OrderedDict()

    def reset(self):
        for option in list(self.options.values()):
            option.reset()

    def setValue(self, key, value):
        if key in self.options:
            self.options[key].setValue(value)
            return True
        else:
            return False  # pragma : no cover - should never happen

    def getValue(self, key, defValue=None):
        if key in self.options:
            return self.options[key].getValue()
        else:
            return defValue

    # For back compatibility
    setOptionValue = setValue
    setSwitchValue = setValue
    getOptionValue = getValue
    getSwitchValue = getValue

    def addSwitch(self, key, *args, **kwargs):
        if key in self.options:
            return False
        self.options[key] = Switch(*args, **kwargs)
        return True

    def addOption(self, key, *args, **kwargs):
        if key in self.options:
            return False
        self.options[key] = TextOption(*args, **kwargs)
        return True

    def setPossibleValues(self, key, possibleValues):
        option = self.options.get(key)
        if option:
            option.setPossibleValues(possibleValues)

    def getOption(self, key):
        return self.options.get(key)

    def getOptionValueMap(self):
        values = {}
        for key, option in list(self.options.items()):
            value = option.getValue()
            if value:
                values[key] = option.getValue()
        return values

    def keys(self):
        return list(self.options.keys())

    def getOptionsForCmdLine(self, *args):
        commandLines = []
        for key, option in list(self.options.items()):
            if self.accept(key, option, *args):
                commandLines.append((key, option.getCmdLineValue()))

        return commandLines

    def moveToEnd(self, keys):
        for key in keys:
            option = self.options.pop(key)
            self.options[key] = option

    def accept(self, key, option, onlyKeys, excludeKeys):
        value = option.getValue()
        if not value or (isinstance(value, str) and value.startswith("<default>")):
            return False

        return (len(onlyKeys) == 0 or key in onlyKeys) and key not in excludeKeys



import sys
import difflib


def _getNumberAt(l, pos):
    start = pos
    eSeen = False
    dotSeen = False
    while start > 0 and l[start-1] in "1234567890.eE-+":
        if l[start-1] in "eE":
            if eSeen:
                break
            eSeen = True
        if l[start-1] == ".":
            if dotSeen:
                break
            dotSeen = True
        start -= 1
    end = pos
    while end < len(l) and l[end] in "1234567890.eE-+":
        if l[end] in "eE":
            if eSeen:
                break
            eSeen = True
        if l[end] == ".":
            if dotSeen:
                break
            dotSeen = True
        end += 1
    return l[start:end], l[end:]


def _fpequalAtPos(l1, l2, tolerance, relTolerance, pos):
    number1, l1 = _getNumberAt(l1, pos)
    number2, l2 = _getNumberAt(l2, pos)
    try:
        equal = False
        deviation = abs(float(number1) - float(number2))
        if tolerance != None and deviation <= tolerance:
            equal = True
        elif relTolerance != None:
            referenceValue = abs(float(number1))
            if referenceValue == 0:
                equal = (deviation == 0)
            elif deviation / referenceValue <= relTolerance:
                equal = True
    except ValueError:
        pass
    return equal, l1, l2


def _fpequal(l1, l2, tolerance, relTolerance):
    pos = 0
    while pos < min(len(l1), len(l2)):
        if l1[pos] != l2[pos]:
            equal, l1, l2 = _fpequalAtPos(l1, l2, tolerance, relTolerance, pos)
            if not equal:
                return False
            pos = 0
        else:
            pos += 1
    if len(l1) == len(l2):
        return True
    else:
        return _fpequalAtPos(l1, l2, tolerance, relTolerance, pos)[0]


def _cmpLines(fromlines, tolines, outlines, tolerance, relTolerance, split):
    for fromline, toline in zip(fromlines, tolines):
        equal = True
        if fromline != toline:
            if split != '':
                fromSplit = fromline.split(split)
                toSplit = toline.split(split)
                if len(fromSplit) == len(toSplit):
                    for f, t in zip(fromSplit, toSplit):
                        if f != t and not _fpequal(f, t, tolerance, relTolerance):
                            equal = False
                            break
                else:
                    equal = False
            elif not _fpequal(fromline, toline, tolerance, relTolerance):
                equal = False
        if equal:
            outlines.write(fromline)
        else:
            outlines.write(toline)


def fpfilter(fromlines, tolines, outlines, tolerance, relTolerance=None, useDifflib=False, split=''):
    if split == 'None':
        split = None
    if not useDifflib:
        _cmpLines(fromlines, tolines, outlines, tolerance, relTolerance, split)
        outlines.writelines(tolines[len(fromlines):])
        return
    s = difflib.SequenceMatcher(None, fromlines, tolines)
    for tag, i1, i2, j1, j2 in s.get_opcodes():
        if tag == "replace" and i2 - i1 == j2 - j1:
            _cmpLines(fromlines[i1:i2], tolines[j1:j2], outlines, tolerance, relTolerance, split)
        else:
            outlines.writelines(tolines[j1:j2])



# Module to find all the child processes of a random process and kill them all (using psutil)

import signal
import os
import subprocess
import psutil


def killProcessAndChildren(pid, sig=None, cmd=None, timeout=5):
    assert pid != os.getpid(), "won't kill myself"
    try:
        parent = psutil.Process(pid)
    except psutil.NoSuchProcess:
        return False
    children = parent.children(recursive=True) + [parent]
    for p in children:
        try:
            p.send_signal(signal.SIGTERM if sig is None else sig)
        except psutil.NoSuchProcess:
            pass # if it's terminated by itself already, so much the better
    _, alive = psutil.wait_procs(children, timeout=timeout)
    if sig is None:
        for p in alive:
            if os.name == "posix":
                p.send_signal(signal.SIGKILL)
            else:
                if cmd is None:
                    cmd = "taskkill /F /PID"
                cmd += " " + str(pid)
                try:
                    subprocess.call(cmd, stdout=open(os.devnull, "w"), stderr=subprocess.STDOUT)
                except OSError:
                    pass
    return True



from gi.repository import Gtk, GObject

import os
import subprocess
import types
import logging
import sys
from . import entrycompletion
from texttestlib import plugins
from .guiutils import guiConfig, SubGUI, GUIConfig, createApplicationEvent
from texttestlib.jobprocess import killProcessAndChildren
from collections import OrderedDict

# The purpose of this class is to provide a means to monitor externally
# started process, so that (a) code can be called when they exit, and (b)
# they can be terminated when TextTest is terminated.


class ProcessTerminationMonitor(plugins.Observable):
    def __init__(self):
        plugins.Observable.__init__(self)
        self.processesForKill = OrderedDict()
        self.exitHandlers = OrderedDict()

    def listQueryKillProcesses(self):
        processesToCheck = guiConfig.getCompositeValue("query_kill_processes", "", modeDependent=True)
        if "all" in processesToCheck:
            processesToCheck = [".*"]
        if len(processesToCheck) == 0:
            return []

        running = []
        triggerGroup = plugins.TextTriggerGroup(processesToCheck)
        for process, description in self.getProcesses():
            if triggerGroup.stringContainsText(description):
                running.append("PID " + str(process.pid) + " : " + description)

        return running

    def getProcesses(self):
        return list(self.processesForKill.values())

    def getProcessIdentifier(self, process):
        # Unfortunately the child_watch_add method needs different ways to
        # identify the process on different platforms...
        if os.name == "posix":
            return process.pid
        else:
            return process._handle

    def startProcess(self, cmdArgs, description="", killOnTermination=True, exitHandler=None, exitHandlerArgs=(), **kwargs):
        process = subprocess.Popen(cmdArgs, stdin=open(os.devnull), **kwargs)
        pidOrHandle = self.getProcessIdentifier(process)
        self.exitHandlers[int(pidOrHandle)] = (exitHandler, exitHandlerArgs)
        if killOnTermination:
            self.processesForKill[int(pidOrHandle)] = (process, description)
        GObject.child_watch_add(pidOrHandle, self.processExited, process.pid)

    def processExited(self, pidOrHandle, condition, pid):
        output = ""
        self.notify("ProcessExited", pid)
        if pidOrHandle in self.processesForKill:
            process = self.processesForKill[pidOrHandle][0]
            if process.stdout is not None:
                output = process.stdout.read().strip()
            del self.processesForKill[pidOrHandle]

        if pidOrHandle in self.exitHandlers:
            exitHandler, exitHandlerArgs = self.exitHandlers.pop(pidOrHandle)
            if exitHandler:
                exitHandler(*exitHandlerArgs)
                if output:
                    command, arg = output.split(" ", 1)
                    self.notify(command, arg)

    def notifyKillProcesses(self, sig=None):
        # Don't leak processes
        if len(self.processesForKill) == 0:
            return
        diag = logging.getLogger("kill processes")
        self.notify("Status", "Terminating all external viewers ...")
        for pid, (process, description) in list(self.processesForKill.items()):
            if pid in self.exitHandlers:
                self.exitHandlers.pop(pid)  # don't call exit handlers in this case, we're terminating
            self.notify("ActionProgress")
            diag.info("Killing '" + description + "' interactive process")
            killProcessAndChildren(process.pid, sig)


processMonitor = ProcessTerminationMonitor()


def openLinkInBrowser(*files):
    if os.name == "nt" and "BROWSER" not in os.environ and len(files) == 1:
        os.startfile(files[0])  # @UndefinedVariable
        createApplicationEvent("the browser to be closed", "browser")
        return 'Started "<default browser> ' + files[0] + '" in background.'
    else:
        browser = os.getenv("BROWSER", "firefox")
        cmdArgs = [browser] + list(files)
        processMonitor.startProcess(cmdArgs, exitHandler=createApplicationEvent,
                                    exitHandlerArgs=("the browser to be closed", "browser"))
        return 'Started "' + " ".join(cmdArgs) + '" in background.'


class GtkActionWrapper:
    def __init__(self):
        self.accelerator = None
        self.diag = logging.getLogger("Interactive Actions")
        title = self.getTitle(includeMnemonics=True)
        actionName = self.getActionName()
        self.gtkAction = Gtk.Action(actionName, title,
                                    self.getTooltip(), self.getStockId())
        self.gtkAction.connect("activate", self.runInteractive)
        if not self.isActiveOnCurrent():
            self.gtkAction.set_property("sensitive", False)

    def getActionName(self):
        return self.getTitle(includeMnemonics=False)

    def getAccelerator(self, title):
        realAcc = guiConfig.getCompositeValue("gui_accelerators", title)
        if realAcc:
            key, mod = Gtk.accelerator_parse(realAcc)
            if Gtk.accelerator_valid(key, mod):
                return realAcc
            else:
                plugins.printWarning("Keyboard accelerator '" + realAcc + "' for action '"
                                     + title + "' is not valid, ignoring ...")

    def addToGroups(self, actionGroup, accelGroup):
        self.accelerator = self._addToGroups(self.gtkAction.get_name().rstrip("."),
                                             self.gtkAction, actionGroup, accelGroup)

    def _addToGroups(self, actionName, gtkAction, actionGroup, accelGroup):
        # GTK 2.12 got fussy about this...
        existingAction = actionGroup.get_action(actionName)
        if existingAction:
            self.diag.info("Removing action with label " + existingAction.get_property("label"))
            actionGroup.remove_action(existingAction)

        accelerator = self.getAccelerator(actionName)
        actionGroup.add_action_with_accel(gtkAction, accelerator)
        gtkAction.set_accel_group(accelGroup)
        gtkAction.connect_accelerator()
        return accelerator

    def setSensitivity(self, newValue):
        self._setSensitivity(self.gtkAction, newValue)

    def _setSensitivity(self, gtkAction, newValue):
        oldValue = gtkAction.get_property("sensitive")
        if oldValue != newValue:
            gtkAction.set_property("sensitive", newValue)


# Introduce an extra level without all the selection-dependent stuff, some actions want
# to inherit from here and it provides a separation
class BasicActionGUI(SubGUI, GtkActionWrapper):
    busy = False

    def __init__(self, *args):
        SubGUI.__init__(self)
        GtkActionWrapper.__init__(self)
        self.topWindow = None

    def checkValid(self, app):
        pass

    def notifyTopWindow(self, window):
        self.topWindow = window

    def getParentWindow(self):
        return self.topWindow

    def isModal(self):
        return True

    def isActiveOnCurrent(self, *args):
        return True

    def getDialogTitle(self):
        return self.getTooltip()

    def createDialog(self):
        if self.isModal():
            dialog = Gtk.Dialog(self.getDialogTitle(), self.getParentWindow(), flags=Gtk.DialogFlags.MODAL)
            dialog.set_modal(True)
        else:
            dialog = Gtk.Dialog(self.getDialogTitle())

        dialog.set_default_response(Gtk.ResponseType.ACCEPT)
        # may need review MB 2018-12-04
        # there may be a flag achieving the behavior, as mentioned here http://docs.adacore.com/live/wave/gtkada/html/gtkada_ug/transition.html
        # dialog.set_has_separator(False)
        return dialog

    def getTitle(self, includeMnemonics=False):
        title = self._getTitle()
        if includeMnemonics:
            return title
        else:
            return title.replace("_", "")

    def getTooltip(self):
        return self.getTitle(includeMnemonics=False)

    def displayInTab(self):
        return False

    def allAppsValid(self):
        return True

    def getStockId(self):
        stockId = self._getStockId()
        if stockId:
            return "gtk-" + stockId

    def _getStockId(self):  # The stock ID for the action, in toolbar and menu.
        pass

    def setObservers(self, observers):
        signals = ["Status", "ActionStart"] + self.getSignalsSent()
        self.diag.info("Observing " + str(self.__class__) + " :")
        for observer in observers:
            for signal in signals:
                if hasattr(observer, "notify" + signal):
                    self.diag.info("-> " + str(observer.__class__))
                    self.addObserver(observer)
                    break

    def getSignalsSent(self):
        return []  # set up like this so every single derived class doesn't have to include it

    def createDialogMessage(self, message, stockIcon):
        buffer = Gtk.TextBuffer()
        buffer.set_text(message)
        textView = Gtk.TextView.new_with_buffer(buffer)
        textView.set_editable(False)
        textView.set_cursor_visible(False)
        textView.set_left_margin(5)
        textView.set_right_margin(5)
        hbox = Gtk.HBox()
        imageBox = Gtk.VBox()
        imageBox.pack_start(Gtk.Image.new_from_stock(stockIcon, Gtk.IconSize.DIALOG), False, True, 0)
        hbox.pack_start(imageBox, False, True, 0)
        scrolledWindow = Gtk.ScrolledWindow()
        # What we would like is that the dialog expands without scrollbars
        # until it reaches some maximum size, and then adds scrollbars. At
        # the moment I cannot make this happen without setting a fixed window
        # size, so I'll set the scrollbar policy to never instead.
        scrolledWindow.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
        scrolledWindow.add(textView)
        scrolledWindow.set_shadow_type(Gtk.ShadowType.IN)
        hbox.pack_start(scrolledWindow, True, True, 0)
        alignment = Gtk.Alignment.new(0.5, 0.5, 1.0, 1.0)
        alignment.set_padding(5, 5, 0, 5)
        alignment.add(hbox)
        return alignment

    def showErrorDialog(self, message):
        self.showErrorWarningDialog(message, Gtk.STOCK_DIALOG_ERROR, "Error")

    def showWarningDialog(self, message):
        self.showErrorWarningDialog(message, Gtk.STOCK_DIALOG_WARNING, "Warning")

    def showErrorWarningDialog(self, message, stockIcon, alarmLevel):
        dialog = self.createAlarmDialog(self.getParentWindow(), message, stockIcon, alarmLevel)
        dialog.add_button(Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT)
        dialog.set_default_response(Gtk.ResponseType.ACCEPT)
        dialog.connect("response", lambda d, r: self._cleanDialog(d))
        dialog.show_all()

    def createAlarmDialog(self, parent, message, stockIcon, alarmLevel):
        dialogTitle = "TextTest " + alarmLevel
        dialog = Gtk.Dialog(dialogTitle, parent, flags=Gtk.DialogFlags.MODAL)
        dialog.set_modal(True)
        # may need review MB 2018-12-05
        # there may be a flag achieving the behavior, as mentioned here http://docs.adacore.com/live/wave/gtkada/html/gtkada_ug/transition.html
        # dialog.set_has_separator(False)

        contents = self.createDialogMessage(message, stockIcon)
        dialog.vbox.pack_start(contents, True, True, 0)
        return dialog

    def showQueryDialog(self, parent, message, stockIcon, alarmLevel, respondMethod, respondData=None):
        dialog = self.createAlarmDialog(parent, message, stockIcon, alarmLevel)
        dialog.set_default_response(Gtk.ResponseType.NO)
        dialog.add_button(Gtk.STOCK_NO, Gtk.ResponseType.NO)
        dialog.add_button(Gtk.STOCK_YES, Gtk.ResponseType.YES)
        if respondMethod:
            dialog.connect("response", respondMethod, respondData)
        dialog.show_all()
        return dialog

    def _cleanDialog(self, dialog, *args):
        entrycompletion.manager.collectCompletions()
        dialog.hide()  # Can't destroy it, we might still want to read stuff from it

    def respond(self, dialog, responseId, *args):
        saidOK = responseId in [Gtk.ResponseType.ACCEPT, Gtk.ResponseType.YES, Gtk.ResponseType.OK]
        try:
            self._respond(saidOK, dialog, *args)
        except plugins.TextTestError as e:
            self.showErrorDialog(str(e))

    def _respond(self, saidOK, dialog, *args):
        if saidOK:
            self._runInteractive()
        else:
            self.cancel()
        if dialog:
            self._cleanDialog(dialog, *args)

    def getConfirmationMessage(self):
        return ""

    def runInteractive(self, *args):
        if self.busy:  # If we're busy with some other action, ignore this one ...
            return

        try:
            confirmationMessage = self.getConfirmationMessage()
            if confirmationMessage:
                self.showQueryDialog(self.getParentWindow(), confirmationMessage,
                                     Gtk.STOCK_DIALOG_WARNING, "Confirmation", self.respond)
            else:
                # Each time we perform an action we collect and save the current registered entries
                # Actions showing dialogs will handle this in the dialog code.
                entrycompletion.manager.collectCompletions()
                self._runInteractive()
        except plugins.TextTestError as e:
            self.showErrorDialog(str(e))
        except Exception as e:
            self.showErrorDialog(plugins.getExceptionString())

    def _runInteractive(self):
        try:
            BasicActionGUI.busy = True
            self.startPerform()
        finally:
            self.endPerform()
            BasicActionGUI.busy = False

    def messageBeforePerform(self):
        # Don't change this by default, most of these things don't take very long
        pass

    def messageAfterPerform(self):
        return "Performed '" + self.getTooltip() + "'."

    def startPerform(self):
        message = self.messageBeforePerform()
        if message != None:
            self.notify("Status", message)
        self.notify("ActionStart")
        self.notify("ActionProgress")
        self.performOnCurrent()
        message = self.messageAfterPerform()
        if message != None:
            self.notify("Status", message)

    def endPerform(self):
        self.notify("ActionStop")

    def cancel(self):
        self.notify("Status", "Action cancelled.")


class ActionGUI(BasicActionGUI):
    busy = False

    def __init__(self, allApps, *args):
        self.currTestSelection = []
        self.currFileSelection = []
        self.currAppSelection = []
        self.validApps = []
        BasicActionGUI.__init__(self)
        for app in allApps:
            self.checkAllValid(app)

    def checkAllValid(self, app):
        for currApp in [app] + app.extras:
            if currApp not in self.validApps:
                self.checkValid(currApp)

    def checkValid(self, app):
        if self.isValidForApp(app):
            self.validApps.append(app)
        else:
            self.diag.info(str(self.__class__) + " invalid for " + repr(app))

    def isValidForApp(self, dummyApp):
        return True

    def shouldShow(self):
        return len(self.validApps) > 0

    def notifyNewTestSelection(self, *args):
        newActive = self.updateSelection(*args)
        self.setSensitivity(newActive)

    def getTestCaseSelection(self):
        testcases = []
        for test in self.currTestSelection:
            for testCase in test.testCaseList():
                if not testCase in testcases:
                    testcases.append(testCase)
        return testcases

    def updateSelection(self, tests, apps, rowCount, *args):
        if rowCount != 1 and self.singleTestOnly():
            self.currTestSelection = []
        else:
            self.currTestSelection = tests
            testClass = self.correctTestClass()
            if testClass:
                self.currTestSelection = [test for test in tests if test.classId() == testClass]

        self.currAppSelection = apps
        newActive = self.allAppsValid() and self.isActiveOnCurrent()
        self.diag.info("New test selection for " + self.getTitle() + "=" +
                       self.describeTests() + " : new active = " + repr(newActive))
        return newActive

    def notifyLifecycleChange(self, test, state, *args):
        newActive = self.isActiveOnCurrent(test, state)
        self.setSensitivity(newActive)

    def notifyNewFileSelection(self, files):
        newActive = self.updateFileSelection(files)
        self.setSensitivity(newActive)

    def updateFileSelection(self, files):
        self.currFileSelection = files
        newActive = self.allAppsValid() and self.isActiveOnCurrent()
        self.diag.info("New file selection for " + self.getTitle() + "=" +
                       repr(files) + " : new active = " + repr(newActive))
        return newActive

    def allAppsValid(self):
        for app in self.currAppSelection:
            if app not in self.validApps:
                self.diag.info("Rejecting due to invalid selected app : " + repr(app))
                return False
        return True

    def isActiveOnCurrent(self, *args):
        return self.shouldShow() and len(self.currTestSelection) > 0

    def singleTestOnly(self):
        return False

    def describeTests(self):
        if len(self.currTestSelection) == 1:
            return "test " + self.currTestSelection[0].getRelPath()
        else:
            return str(len(self.currTestSelection)) + " tests"

    def correctTestClass(self):
        pass

    def messageAfterPerform(self):
        return "Performed '" + self.getTooltip() + "' on " + self.describeTests() + "."

    def createTextWidget(self, name, scroll=False):
        frame = Gtk.Frame()
        frame.set_shadow_type(Gtk.ShadowType.OUT)
        frame.set_border_width(1)
        view = Gtk.TextView()
        view.set_wrap_mode(Gtk.WrapMode.WORD)
        view.set_name(name)
        buffer = view.get_buffer()
        if scroll:
            view = self.addScrollBars(view, Gtk.PolicyType.AUTOMATIC)
        frame.add(view)
        return frame, buffer

    def tryResize(self, dialog):
        horfrac, verfrac = self.getSizeAsWindowFraction()
        if horfrac is not None and self.topWindow is not None:
            width, height = self.topWindow.get_size()
            dialog.resize(int(width * horfrac), int(height * verfrac))

    def getSizeAsWindowFraction(self):
        return None, None

    def createButton(self):
        return self._createButton(self.gtkAction, self.getTooltip())

    def _createButton(self, action, tooltip):
        button = Gtk.Button()
        # needs review MB 2018-12-05
        button.set_related_action(action)
        # In theory all this should be automatic, but it appears not to work
        if self.getStockId():
            image = Gtk.Image.new_from_stock(self.getStockId(), Gtk.IconSize.BUTTON)
            button.set_image(image)
            image.show()

        button.set_tooltip_text(tooltip)
        button.show()
        return button

    def findTestsToReload(self):
        tests = []
        for test in self.currTestSelection:
            for currTest in test.getAllTestsToRoot():
                if currTest not in tests:
                    currTest.refreshFiles()
                    if currTest.hasLocalConfig():
                        tests.append(currTest)
        return tests

    def reloadConfigForSelected(self):
        for appOrTest in self.currAppSelection + self.findTestsToReload():
            self.notify("Status", "Rereading configuration for " + repr(appOrTest) + " ...")
            self.notify("ActionProgress")
            appOrTest.reloadConfiguration()


# These actions consist of bringing up a dialog and only doing that
# (i.e. the dialog is not a mechanism to steer how the action should be run)
class ActionResultDialogGUI(ActionGUI):
    def __init__(self, *args, **kw):
        self.dialog = None
        ActionGUI.__init__(self, *args, **kw)

    def performOnCurrent(self):
        self.dialog = self.createDialog()
        self.addContents()
        self.createButtons()
        self.tryResize(self.dialog)
        self.dialog.show_all()

    def addContents(self):  # pragma: no cover - documentation only
        pass

    def createButtons(self):
        self.dialog.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.ACCEPT)
        self.dialog.set_default_response(Gtk.ResponseType.ACCEPT)
        self.dialog.connect("response", self.respond)

    def respond(self, dialog, responseId):
        if responseId != Gtk.ResponseType.NONE:
            self._cleanDialog(dialog)


class ComboBoxListFinder:
    def __init__(self, combobox):
        self.model = combobox.get_model()
        self.textColumn = combobox.get_entry_text_column()

    def __call__(self):
        entries = []
        self.model.foreach(self.getText, entries)
        return entries

    def getText(self, model, dummyPath, iter, entries):
        text = model.get_value(iter, self.textColumn)
        entries.append(text)


# A utility class to set and get the indices of options in radio button groups.
class RadioGroupIndexer:
    def __init__(self, listOfButtons):
        self.buttons = listOfButtons

    def getActiveIndex(self):
        for i in range(0, len(self.buttons)):
            if self.buttons[i].get_active():
                return i

    def setActiveIndex(self, index):
        self.buttons[index].set_active(True)


class OptionGroupGUI(ActionGUI):
    def __init__(self, *args):
        ActionGUI.__init__(self, *args)
        self.groupBoxes = {}
        self.optionGroup = plugins.OptionGroup(self.getTabTitle())
        # convenience shortcuts...
        self.addOption = self.optionGroup.addOption
        self.addSwitch = self.optionGroup.addSwitch

    def updateOptions(self):
        return False

    def updateForConfig(self, option):
        fromConfig = guiConfig.getCompositeValue("gui_entry_overrides", option.name)
        # only do this if it hasn't previously been manually overwritten
        if fromConfig is not None and fromConfig != "<not set>" and option.getValue() == option.defaultValue:
            newValue = fromConfig
            if isinstance(option.getValue(), int):
                newValue = int(fromConfig)
            option.setValue(newValue)
            return fromConfig

    def createLabelEventBox(self, option, separator):
        label = Gtk.EventBox()
        label.add(Gtk.Label(label=option.name + separator))
        if option.description and type(option.description) in (str, bytes):
            label.set_tooltip_text(option.description)
        return label

    def destroyedEntry(self, entry, data):
        option, entryOrBuffer = data
        if hasattr(option.valueMethod, "__self__") and option.valueMethod.__self__ is entryOrBuffer:
            option.setMethods(None, None)

    def connectEntry(self, option, entryOrBuffer, entryWidget):
        entryWidget.connect("destroy", self.destroyedEntry, (option, entryOrBuffer))

        def setText(t):
            entryOrBuffer.set_text(str(t))
        setText(option.getValue())
        # Don't pass entry.set_text directly, it will mess up StoryText's programmatic method interception
        option.setMethods(self.getGetTextMethod(entryOrBuffer), setText)
        if option.changeMethod:
            entryOrBuffer.connect("changed", option.changeMethod)

    def getGetTextMethod(self, widget):
        if isinstance(widget, Gtk.SpinButton):
            # Would be nice to return widget.get_value_as_int but that returns the wrong answer from
            # dialogs that have been closed
            def get_text():
                text = widget.get_text()
                return float(text) if widget.get_digits() else int(text)
            return get_text
        elif isinstance(widget, Gtk.Entry):
            return widget.get_text
        else:
            def get_text():
                return widget.get_text(widget.get_start_iter(), widget.get_end_iter(), True)
            return get_text

    def addValuesFromConfig(self, option, includeOverrides=True):
        if includeOverrides:
            newValue = self.updateForConfig(option)
            if newValue:
                option.addPossibleValue(newValue)
        for extraOption in self.getConfigOptions(option):
            option.addPossibleValue(extraOption)

    def createRadioButtonCollection(self, switch, optionGroup):
        hbox = Gtk.HBox()
        if len(switch.name) > 0:
            label = self.createLabelEventBox(switch, ":")
            hbox.pack_start(label, False, False, 0)
        for button in self.createRadioButtons(switch, optionGroup):
            hbox.pack_start(button, True, False, 0)
        hbox.show_all()
        return hbox

    def setConfigOverride(self, switch, index, option, *args):
        configName = self.getNaming(switch.name, option, *args)
        if index == switch.getValue() or guiConfig.getCompositeValue("gui_entry_overrides", configName) == "1":
            switch.setValue(index)

    def getNaming(self, switchName, option, *args):
        if len(switchName) > 0:
            return switchName + ":" + option
        else:
            return option

    def createRadioButtons(self, switch, optionGroup):
        buttons = []
        mainRadioButton = None
        individualToolTips = type(switch.description) == list
        for index, option in enumerate(switch.options):
            self.setConfigOverride(switch, index, option, optionGroup)
            radioButton = Gtk.RadioButton.new_with_mnemonic_from_widget(mainRadioButton, option)
            self.setRadioButtonName(radioButton, option, optionGroup)
            if individualToolTips:
                radioButton.set_tooltip_text(switch.description[index])

            buttons.append(radioButton)
            if not mainRadioButton:
                mainRadioButton = radioButton
            if switch.getValue() == index:
                radioButton.set_active(True)
            else:
                radioButton.set_active(False)

        indexer = RadioGroupIndexer(buttons)
        switch.setMethods(indexer.getActiveIndex, indexer.setActiveIndex)
        return buttons

    def createFrame(self, group, name):
        frame = Gtk.Frame.new(name)
        frame.set_label_align(0.5, 0.5)
        frame.set_shadow_type(Gtk.ShadowType.IN)
        frame.add(self.createGroupBox(group))
        return frame

    def createGroupBox(self, group):
        frameBox = Gtk.VBox()
        frameBox.set_border_width(10)
        self.fillVBox(frameBox, group)
        self.groupBoxes[group] = frameBox
        return frameBox

    def setGroupSensitivity(self, group, *args, **kw):
        widget = self.groupBoxes.get(group)
        self.setChildSensitivity(widget, *args, **kw)

    def setChildSensitivity(self, widget, sensitive, ignoreWidget=None):
        if widget is ignoreWidget or isinstance(widget, Gtk.RadioButton):
            return
        # needs review, just replaced ComboBoxEntry by ComboBox, maybe ComboBoxText would be better? MB 2018-12-05
        elif isinstance(widget, (Gtk.Entry, Gtk.CheckButton, Gtk.ComboBox)):
            widget.set_sensitive(sensitive)
        elif hasattr(widget, "get_children"):
            for child in widget.get_children():
                self.setChildSensitivity(child, sensitive, ignoreWidget)

    def setRadioButtonName(self, *args):
        pass  # Don't bother by default, it's easy to set stupid names...

    def createComboBox(self, switch, *args):
        combobox = Gtk.ComboBoxText()
        combobox.set_name(switch.name)
        for index, option in enumerate(switch.options):
            combobox.append_text(option)
            self.setConfigOverride(switch, index, option, *args)
            if switch.getValue() == index:
                combobox.set_active(index)

        switch.setMethods(combobox.get_active, combobox.set_active)
        box = Gtk.VBox()
        box.pack_start(Gtk.Label(""), True, True, 0)
        box.pack_start(combobox, True, True, 0)
        return box

    def transferEnable(self, enabler, switch):
        if enabler.get_active():
            switch.updateMethod(True)

    def extractSwitches(self, optionGroup):
        options, switches = [], []
        for option in list(optionGroup.options.values()):
            if isinstance(option, plugins.Switch):
                switches.append(option)
            else:
                options.append(option)
        return options, switches

    def findAutoEnableInfo(self, switches):
        info = {}
        for switch in switches:
            if isinstance(switch, plugins.Switch) and switch.autoEnable:
                for enableSwitchName in switch.autoEnable:
                    enabler = self.getOption(enableSwitchName)
                    if enabler:
                        info[enabler] = switch
        return info

    def createSwitchWidget(self, switch, optionGroup, autoEnableInfo):
        if len(switch.options) >= 1:
            if switch.hideOptions:
                return self.createComboBox(switch, optionGroup)
            else:
                return self.createRadioButtonCollection(switch, optionGroup)
        else:
            return self.createCheckBox(switch, autoEnableInfo)

    def createCheckBox(self, switch, autoEnableInfo):
        self.updateForConfig(switch)
        checkButton = Gtk.CheckButton(switch.name)
        if switch.description:
            checkButton.set_tooltip_text(switch.description)

        if int(switch.getValue()):
            checkButton.set_active(True)
        # Don't pass checkButton.set_active as that will screw up StoryText's interception of it
        switch.setMethods(checkButton.get_active, lambda x: checkButton.set_active(x))
        if switch in autoEnableInfo:
            toEnable = autoEnableInfo.get(switch)
            checkButton.connect("toggled", self.transferEnable, toEnable)

        checkButton.show()
        return checkButton

    def createComboBoxEntry(self, option):
        # may need review MB 2018-12-04
        combobox = Gtk.ComboBoxText.new_with_entry()
        entry = combobox.get_child()
        combobox.set_row_separator_func(self.isRowSeparator)
        option.setPossibleValuesMethods(combobox.append_text, ComboBoxListFinder(combobox))

        option.setClearMethod(combobox.get_model().clear)
        return combobox, entry

    def isRowSeparator(self, model, iter):
        text = model.get_value(iter, 0)
        return text == "-" * 10

    def createOptionWidget(self, option):
        optionName = option.name.strip()
        if option.multilineEntry:
            return self.createTextWidget(optionName)
        else:
            box = Gtk.HBox()
            value = option.getValue()
            if isinstance(value, int) or isinstance(value, float):
                adjustment = Gtk.Adjustment(value=value, lower=option.minimum,
                                            upper=option.maximum, step_incr=1)
                digits = int(isinstance(value, float))
                widget = Gtk.SpinButton.new(adjustment, 0., digits)
                widget.set_numeric(True)
                entry = widget
            elif option.usePossibleValues():
                widget, entry = self.createComboBoxEntry(option)
                widget.set_name(optionName + " (Combo Box)")
            else:
                widget = Gtk.Entry()
                entry = widget
            box.pack_start(widget, True, True, 0)
            entry.set_name(optionName)
            if not isinstance(widget, Gtk.SpinButton):
                entrycompletion.manager.register(entry)
            # Options in drop-down lists don't change, so we just add them once and for all.
            for text in option.listPossibleValues():
                # needs review, just added str to work around expected type error, MB 2018-12-05
                entrycompletion.manager.addTextCompletion(str(text))

            return box, entry

    def createFileChooserDialog(self, box, entry, option):
        button = Gtk.Button("...")
        box.pack_start(button, False, False, 0)
        button.connect("clicked", self.showFileChooser, entry, option)

    def showFileChooser(self, dummyWidget, entry, option):
        dialog = Gtk.FileChooserDialog("Select a file",
                                       self.getParentWindow(),
                                       Gtk.FileChooserAction.OPEN,
                                       (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
        self.startFileChooser(dialog, entry, option)

    def startFileChooser(self, dialog, entry, option):
        # Folders is a list of pairs (short name, absolute path),
        # where 'short name' means the name given in the config file, e.g.
        # 'temporary_filter_files' or 'filter_files' ...
        dialog.set_modal(True)
        folders, defaultFolder = option.getDirectories()
        dialog.connect("response", self.respondFileChooser, entry)
        # If current entry forms a valid path, set that as default
        currPath = entry.get_text()
        currDir = os.path.split(currPath)[0]
        if os.path.isdir(currDir):
            dialog.set_current_folder(currDir)
        elif defaultFolder and os.path.isdir(os.path.abspath(defaultFolder)):
            dialog.set_current_folder(os.path.abspath(defaultFolder))
        for folder in folders:
            dialog.add_shortcut_folder(folder)
        dialog.set_default_response(Gtk.ResponseType.OK)
        dialog.show()

    def respondFileChooser(self, dialog, response, entry):
        if response == Gtk.ResponseType.OK:
            entry.set_text(dialog.get_filename().replace("\\", "/"))
            entry.set_position(-1)  # Sets position last, makes it possible to see the vital part of long paths
        dialog.destroy()

    def getConfigOptions(self, option):
        fromConfig = guiConfig.getCompositeValue("gui_entry_options", option.name)
        if fromConfig is None:  # Happens on initial startup with no apps...
            return []
        return fromConfig

    def getCommandLineArgs(self, optionGroup, onlyKeys=[], excludeKeys=[]):
        args = []
        for key, value in optionGroup.getOptionsForCmdLine(onlyKeys, excludeKeys):
            args.append("-" + key)
            if value:
                args.append(str(value))
        return args

    def hasPerformance(self, apps, *args):
        for app in apps:
            if app.hasPerformance(*args):
                return True
        return False

    def addApplicationOptions(self, allApps, optionGroup, inputOptions={}):
        if len(allApps) > 0:
            for app in allApps:
                app.addToOptionGroups(allApps, [optionGroup])
        else:
            configObject = self.makeDefaultConfigObject(inputOptions)
            configObject.addToOptionGroups(allApps, [optionGroup])

    def makeDefaultConfigObject(self, inputOptions):
        return plugins.importAndCall("default", "getConfig", inputOptions)


class ActionTabGUI(OptionGroupGUI):
    def __init__(self, *args):
        OptionGroupGUI.__init__(self, *args)
        self.diag.info("Creating action tab for " + self.getTabTitle() +
                       ", sensitive " + repr(self.shouldShowCurrent()))
        self.vbox = Gtk.VBox()

    def shouldShowCurrent(self, *args):
        return self.gtkAction.get_property("sensitive")

    def createView(self):
        self.fillVBox(self.vbox, self.optionGroup)
        self.createButtons(self.vbox)
        self.vbox.show_all()
        self.widget = self.addScrollBars(self.vbox, hpolicy=Gtk.PolicyType.AUTOMATIC)
        self.widget.set_name(self.getTabTitle() + " Tab")
        return self.widget

    def setSensitivity(self, newValue):
        ActionGUI.setSensitivity(self, newValue)
        self.diag.info("Sensitivity of " + self.getTabTitle() + " changed to " + repr(newValue))
        if self.shouldShowCurrent():
            self.updateOptions()

    def displayInTab(self):
        return True

    def notifyReset(self, *args):
        self.optionGroup.reset()

    def fillVBox(self, vbox, optionGroup):
        options, switches = self.extractSwitches(optionGroup)
        if len(options) > 0:
            # Creating 0-row table gives a warning ...
            table = Gtk.Table(len(options), 2, homogeneous=False)
            table.set_row_spacings(1)
            rowIndex = 0
            for option in options:
                self.addValuesFromConfig(option)

                labelEventBox = self.createLabelEventBox(option, separator="  ")
                labelEventBox.get_children()[0].set_alignment(1.0, 0.5)
                table.attach(labelEventBox, 0, 1, rowIndex, rowIndex + 1, xoptions=Gtk.AttachOptions.FILL, xpadding=1)
                entryWidget, entryOrBuffer = self.createOptionWidget(option)
                self.connectEntry(option, entryOrBuffer, entryWidget)
                if isinstance(entryOrBuffer, Gtk.Entry):
                    entryOrBuffer.connect("activate", self.runInteractive)
                table.attach(entryWidget, 1, 2, rowIndex, rowIndex + 1)
                rowIndex += 1
                table.show_all()
            vbox.pack_start(table, False, False, 0)

        autoEnableInfo = self.findAutoEnableInfo(switches)
        for switch in switches:
            widget = self.createSwitchWidget(switch, optionGroup, autoEnableInfo)
            vbox.pack_start(widget, False, False, 0)

    def createResetButton(self):
        button = Gtk.Button("Reset Tab")
        button.set_name("Reset " + self.getTabTitle() + " Tab")
        button.connect("clicked", self.notifyReset)
        button.set_tooltip_text("Reset all the settings in the current tab to their default values")
        button.show()
        return button

    def createButtons(self, vbox):
        self.addCentralButton(vbox, self.createButton(), padding=8)
        self.addCentralButton(vbox, self.createResetButton(), padding=16)

    def addCentralButton(self, vbox, button, padding):
        buttonbox = Gtk.HButtonBox()
        buttonbox.pack_start(button, True, False, 0)
        vbox.pack_start(buttonbox, False, False, padding)

    def createOptionWidget(self, option):
        box, entry = OptionGroupGUI.createOptionWidget(self, option)
        if option.selectFile:
            self.createFileChooserDialog(box, entry, option)
        return (box, entry)


class ActionDialogGUI(OptionGroupGUI):
    def runInteractive(self, *args):
        if self.busy:  # If we're busy with some other action, ignore this one ...
            return

        self.updateOptions()
        try:
            self.showConfigurationDialog()
        except plugins.TextTestError as e:
            self.showErrorDialog(str(e))

    def showConfigurationDialog(self):
        dialog = self.createDialog()
        alignment = self.createAlignment()
        vbox = Gtk.VBox()
        fileChooser, fileChooserOption = self.fillVBox(vbox, self.optionGroup)
        if self.needsScrollBars():
            vbox = self.addScrollBars(vbox, Gtk.PolicyType.AUTOMATIC)
        alignment.add(vbox)
        dialog.vbox.pack_start(alignment, True, True, 0)
        self.createButtons(dialog, fileChooser, fileChooserOption)
        self.tryResize(dialog)
        dialog.show_all()
        return dialog

    def needsScrollBars(self):
        return False

    def getConfirmationDialogSettings(self):
        return Gtk.STOCK_DIALOG_WARNING, "Confirmation"

    def _respond(self, saidOK=True, dialog=None, fileChooserOption=None):
        if saidOK:
            try:
                message = self.getConfirmationMessage()
                if message:
                    stockId, level = self.getConfirmationDialogSettings()
                    self.showQueryDialog(self.getQueryParentWindow(dialog), message, stockId, level,
                                         self.confirmationRespond, fileChooserOption)
                else:
                    self.defaultRespond(saidOK, dialog, fileChooserOption)
            except plugins.TextTestError as e:
                self.showErrorDialog(str(e))
            except Exception:
                self.showErrorDialog(plugins.getExceptionString())
        else:
            self.defaultRespond(saidOK, dialog, fileChooserOption)

    def getQueryParentWindow(self, dialog):
        if dialog:
            return dialog
        else:
            return self.getParentWindow()

    def confirmationRespond(self, dialog, responseId, fileChooserOption):
        saidOK = responseId == Gtk.ResponseType.YES
        self.defaultRespond(saidOK, dialog)
        if saidOK:
            parent = dialog.get_transient_for()
            if isinstance(parent, Gtk.Dialog):
                self._cleanDialog(parent, fileChooserOption)

    def _cleanDialog(self, dialog, fileChooserOption=None):
        if fileChooserOption:
            fileChooserOption.resetDefault()  # Must do this, because we can't rely on reading from invisible FileChoosers
        OptionGroupGUI._cleanDialog(self, dialog)

    def defaultRespond(self, *args):
        OptionGroupGUI._respond(self, *args)

    def createAlignment(self):
        alignment = Gtk.Alignment.new(1.0, 1.0, 1.0, 1.0)
        alignment.set_padding(5, 5, 5, 5)
        return alignment

    def getOkStock(self, scriptName):
        if scriptName.startswith("load"):
            return "texttest-stock-load"
        elif scriptName.startswith("save"):
            return Gtk.STOCK_SAVE
        else:
            return Gtk.STOCK_OK

    def createButtons(self, dialog, fileChooser, fileChooserOption):
        dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        actionScriptName = self.getTooltip()
        dialog.add_button(self.getOkStock(actionScriptName.lower()), Gtk.ResponseType.ACCEPT)
        dialog.set_default_response(Gtk.ResponseType.ACCEPT)
        if fileChooser:
            fileChooser.connect("file-activated", self.simulateResponse, dialog)
            # Don't pass set_filename directly, will interfere with StoryText's attempts to intercept it
            fileChooserOption.setMethods(fileChooser.get_filename, lambda f: fileChooser.set_filename(f))

        dialog.connect("response", self.respond, fileChooserOption)

    def simulateResponse(self, dummy, dialog):
        dialog.response(Gtk.ResponseType.ACCEPT)

    def getOrderedOptions(self, optionGroup):
        return list(optionGroup.options.values())

    def fillVBox(self, vbox, optionGroup, includeOverrides=True):
        fileChooser, fileChooserOption = None, None
        allOptions = self.getOrderedOptions(optionGroup)
        autoEnableInfo = self.findAutoEnableInfo(allOptions)
        for option in allOptions:
            self.addValuesFromConfig(option, includeOverrides)

            if isinstance(option, plugins.Switch):
                widget = self.createSwitchWidget(option, optionGroup, autoEnableInfo)
                vbox.pack_start(widget, False, False, 0)
            elif option.selectFile or option.selectDir or option.saveFile:
                if not self.showFileChooserAsDialog():
                    fileChooserOption = option
                    fileChooser = self.createFileChooser(option)
                    if len(allOptions) > 1 and not option.saveFile:
                        # If there is other stuff, add a frame round the file chooser so we can see what it's for
                        # Don't do this when approving as it shouldn't be necessary
                        labelEventBox = self.createLabelEventBox(option, separator=":")
                        frame = Gtk.Frame()
                        frame.set_label_widget(labelEventBox)
                        frame.add(fileChooser)
                        vbox.pack_start(frame, True, True, 0)
                    else:
                        vbox.pack_start(fileChooser, True, True, 0)
                else:
                    widget, entry = self.createOptWidget(vbox, option)
                    self.createFileChooserDialog(widget, entry, option)
            else:
                self.createOptWidget(vbox, option)

        if fileChooser:
            # File choosers seize the focus, mostly we want to get it back and put it on the first text entry
            self.set_focus(vbox)
        return fileChooser, fileChooserOption

    def set_focus(self, vbox):
        for child in vbox.get_children():
            if isinstance(child, Gtk.Container) and not isinstance(child, Gtk.FileChooser):
                for gchild in child.get_children():  # This may cause indeterministic behavior if called on FileChoosers, see TTT-2485
                    if isinstance(gchild, Gtk.Entry) or isinstance(gchild, Gtk.ComboBox):
                        gchild.get_toplevel().connect("map", lambda x: gchild.grab_focus())
                        return

    def createOptWidget(self, vbox, option):
        labelEventBox = self.createLabelEventBox(option, separator=":")
        self.addLabel(vbox, labelEventBox)
        entryWidget, entryOrBuffer = self.createOptionWidget(option)
        if isinstance(entryOrBuffer, Gtk.Entry):
            entryOrBuffer.set_activates_default(True)
            vbox.pack_start(entryWidget, False, False, 0)
        else:
            vbox.pack_start(entryWidget, True, True, 0)
        self.connectEntry(option, entryOrBuffer, entryWidget)
        return entryWidget, entryOrBuffer

    def addLabel(self, vbox, label):
        hbox = Gtk.HBox()
        hbox.pack_start(label, False, False, 0)
        vbox.pack_start(hbox, False, False, 2)

    def createRadioButtonCollection(self, switch, optionGroup):
        if optionGroup is not self.optionGroup:
            # If we're not part of the main group, we've got a frame already, store horizontally in this case
            return OptionGroupGUI.createRadioButtonCollection(self, switch, optionGroup)

        if switch.name:
            frame = Gtk.Frame.new(switch.name)
        else:
            frame = Gtk.Frame()
        frameBox = Gtk.VBox()
        for button in self.createRadioButtons(switch, optionGroup):
            frameBox.pack_start(button, True, True, 0)
        frame.add(frameBox)
        return frame

    def showFileChooserAsDialog(self):
        return False

    def getFileChooserFlag(self, option):
        if option.selectFile:
            return Gtk.FileChooserAction.OPEN
        elif option.selectDir:
            return Gtk.FileChooserAction.SELECT_FOLDER
        else:
            return Gtk.FileChooserAction.SAVE

    def createFileChooser(self, option):
        fileChooser = Gtk.FileChooserWidget.new(self.getFileChooserFlag(option))
        fileChooser.set_name("File Chooser for '" + self.getTooltip() + "'")
        # The following line has no effect until this GTK bug is fixed...
        # https://bugzilla.gnome.org/show_bug.cgi?id=440667
        fileChooser.set_show_hidden(True)
        folders, defaultFolder = option.getDirectories()
        if option.selectDir and option.getValue():
            startFolder = option.getValue()
        elif defaultFolder and os.path.isdir(os.path.abspath(defaultFolder)):
            startFolder = os.path.abspath(defaultFolder)
        else:
            startFolder = os.getcwd()  # Just to make sure we always have some dir ...

        # We want a filechooser dialog to let the user choose where, and
        # with which name, to save the selection.
        fileChooser.set_current_folder(startFolder)
        for folder in folders:
            try:
                fileChooser.add_shortcut_folder(folder)
            except GObject.GError:
                pass  # Get this if the folder is already added, e.g. if it's the home directory

        if option.selectFile and self.set_filename_is_working():
            if option.getValue():
                fileChooser.set_filename(option.getValue())
            else:
                value = self.getDefaultFileChooserValue(startFolder)
                if value:
                    fileChooser.set_filename(value)

        fileChooser.set_local_only(True)
        return fileChooser

    def set_filename_is_working(self):
        # A particularly nasty bug was introduced in GTK 2.20, and fixed in GTK 2.24
        # https://bugzilla.gnome.org/show_bug.cgi?id=643170
        # It not only prevents set_filename from working but causes GTK events to be created and never executed
        # making the GUI think it is constantly busy and sending our throbber into an infinite loop
        # See bugs TTT-3469, TTT-3483
        # this block can probably removed MB 2018-12-04
        # if Gtk.gtk_version < (2, 20) or Gtk.gtk_version >= (2, 24):
        #    return True

        fileChooserConfigFile = self.getFileChooserConfigFile()
        if not self.showHiddenFilesEnabled(fileChooserConfigFile):
            return True

        sys.stderr.write("WARNING: unable to select files in File Chooser, due to a GTK+ bug.\n" +
                         "Either upgrade GTK+ to at least version 2.24, or set 'ShowHidden=false' in your configuration file at\n" +
                         fileChooserConfigFile + "\n")
        return False

    def getFileChooserConfigFile(self):
        xdgConfigHome = os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config"))
        return os.path.join(xdgConfigHome, "gtk-2.0", "gtkfilechooser.ini")

    def showHiddenFilesEnabled(self, configFile):
        if os.path.isfile(configFile):
            with open(configFile) as f:
                for line in f:
                    if "ShowHidden" in line and "true" in line:
                        return True
        return False

    def getDefaultFileChooserValue(self, startFolder):
        appSuffices = ["." + app.name for app in self.validApps]
        for f in sorted(os.listdir(startFolder)):
            path = os.path.join(startFolder, f)
            if not self.onlyDataFilesInFileChooser() or (os.path.isfile(path) and not any((suffix in f for suffix in appSuffices))):
                return path

    def onlyDataFilesInFileChooser(self):
        return False

    def getFilterFileDirs(self, allApps, **kw):
        if len(allApps) > 0:
            return allApps[0].getFilterFileDirectories(allApps, **kw)
        else:
            return []


class CloseWindowCancelException(plugins.TextTestException):
    pass


class InteractiveActionConfig:
    def getColourDictionary(self):
        return GUIConfig.getDefaultColours()

    def getDefaultAccelerators(self):
        return {}

    def getReplacements(self):
        # Return a dictionary mapping classes above to what to replace them with
        return {}




"""
Actions for managing selections and filterings of the test tree
"""
from gi.repository import Gtk
import os
import operator
import logging
from texttestlib import plugins
from .. import guiplugins
from functools import reduce


class AllTestsHandler:
    def __init__(self):
        self.rootTestSuites = []

    def addSuites(self, suites):
        self.rootTestSuites += suites

    def getTests(self, suite):
        if suite.classId() == "test-suite" and len(suite.testcases):
            return reduce(operator.add, (self.getTests(subsuite) for subsuite in suite.testcases), [])
        else:
            # include empty suites here, if they match
            return [suite]

    def findAllTests(self):
        return reduce(operator.add, (self.getTests(suite) for suite in self.rootTestSuites), [])

    def findTestsNotIn(self, tests):
        return [test for test in self.findAllTests() if test not in tests]


class SelectTests(guiplugins.ActionTabGUI, AllTestsHandler):
    def __init__(self, allApps, dynamic, *args):
        guiplugins.ActionTabGUI.__init__(self, allApps)
        AllTestsHandler.__init__(self)
        self.dynamic = dynamic
        self.filterAccel = None
        self.filterAction = Gtk.Action("Filter", "Filter",
                                       self.getFilterTooltip(), self.getStockId())
        self.filterAction.connect("activate", self.filterTests)
        self.selectDiag = logging.getLogger("Select Tests")
        self.addOption("vs", "Tests for version", description="Select tests for a specific version.",
                       possibleValues=self.getPossibleVersions(allApps))
        self.selectionGroup = plugins.OptionGroup(self.getTabTitle())
        self.selectionGroup.addSwitch("select_in_collapsed_suites", "Select in collapsed suites",
                                      0, description="Select in currently collapsed suites as well?")
        currSelectDesc = ["Unselect all currently selected tests before applying the new selection criteria.",
                          "Apply the new selection criteria only to the currently selected tests, to obtain a subselection.",
                          "Keep the currently selected tests even if they do not match the new criteria, and extend the selection with all other tests which meet the new criteria.",
                          "After applying the new selection criteria to all tests, unselect the currently selected tests, to exclude them from the new selection."]
        self.selectionGroup.addSwitch("current_selection", options=["Discard", "Refine", "Extend", "Exclude"],
                                      description=currSelectDesc)
        self.filteringGroup = plugins.OptionGroup("Filtering")
        currFilterDesc = ["Show all tests which match the criteria, and hide all those that do not.",
                          "Hide all tests which do not match the criteria. Do not show any tests that aren't already shown.",
                          "Show all tests which match the criteria. Do not hide any tests that are currently shown."]
        self.filteringGroup.addSwitch("current_filtering", options=[
                                      "Discard", "Refine", "Extend"], description=currFilterDesc)
        excludeKeys = set(self.optionGroup.keys())  # remember these so we don't try and save them to selections
        self.addApplicationOptions(allApps, self.optionGroup)
        appNames = set([app.name for app in allApps])
        if len(appNames) > 1:
            self.addOption("app", "App names containing",
                           description="Select tests for which the application name matches the entered text. The text can be a regular expression.")
        if self.hasPerformance(allApps, "performance"):
            self.optionGroup.moveToEnd(["r"])
            self.addOption("fastest", "Only fastest tests", 0,
                           description="Select the given number of tests which match the other selected criteria and have the fastest runtimes")
            self.addOption("slowest", "Only slowest tests", 0,
                           description="Select the given number of tests which match the other selected criteria and have the longest runtimes")
        self.optionGroup.moveToEnd(["grep", "grepfile"])
        if self.dynamic:
            self.addSwitch("std", options=["Use test-files from current run", "Use stored test-files"], description=["When searching using 'test-files containing', look in the results of tests in the current run",
                                                                                                                     "When searching using 'test-files containing', look in the stored results, i.e. the same search as would be done in the static GUI"])
        self.appKeys = set(self.optionGroup.keys())
        self.appKeys.difference_update(excludeKeys)

    def addToGroups(self, actionGroup, accelGroup):
        guiplugins.ActionTabGUI.addToGroups(self, actionGroup, accelGroup)
        self.filterAccel = self._addToGroups("Filter", self.filterAction, actionGroup, accelGroup)

    def notifyAllRead(self, *args):
        allStems = self.findAllStems()
        defaultTestFile = self.findDefaultTestFile(allStems)
        self.notify("AllStems", allStems, defaultTestFile)
        self.optionGroup.setValue("grepfile", defaultTestFile)
        self.optionGroup.setPossibleValues("grepfile", allStems)

    def setRadioButtonName(self, radioButton, option, optionGroup):
        radioButton.set_name(option + " for " + optionGroup.name)

    def findDefaultTestFile(self, allStems):
        for app in self.validApps:
            logFile = app.getConfigValue("log_file")
            if len(allStems) == 0 or logFile in allStems:
                return logFile
        if len(allStems) > 0:
            return allStems[0]
        else:
            return "none"

    def findAllStems(self):
        stems = set()
        defStems = set()
        importantStems = set()
        for suite in self.rootTestSuites:
            defStems.update(suite.defFileStems())
            importantStems.update(suite.getCompositeConfigValue("gui_entry_options", "test-file_to_search"))
            exclude = suite.getDataFileNames() + ["file_edits"]

            def predicate(stem, vset): return stem not in exclude and (stem in defStems or len(vset) > 0)
            for test in suite.testCaseList():
                stems.update(test.dircache.findAllStems(predicate))
        defStems.intersection_update(stems)
        stems.difference_update(defStems)
        self.selectDiag.info("Found important stems " + repr(importantStems))
        stems.difference_update(importantStems)
        subLists = [importantStems, defStems, stems]
        if self.dynamic:
            subLists.insert(0, ["free_text"])
        return self.makeStemList(subLists)

    def makeStemList(self, subLists):
        separator = "-" * 10
        allStems = sorted(subLists[0])
        for extraList in subLists[1:]:
            if len(allStems) and len(extraList):
                allStems.append(separator)
            allStems += sorted(extraList)
        return allStems

    def getPossibleVersions(self, allApps):
        possVersions = []
        for app in allApps:
            for possVersion in self._getPossibleVersions(app):
                if possVersion not in possVersions:
                    possVersions.append(possVersion)
        return possVersions

    def _getPossibleVersions(self, app):
        fullVersion = app.getFullVersion()
        extraVersions = app.getExtraVersions()
        if len(fullVersion) == 0:
            return ["<default>"] + extraVersions
        else:
            return [fullVersion] + [fullVersion + "." + extra for extra in extraVersions]

    def isActiveOnCurrent(self, *args):
        return True

    def getSignalsSent(self):
        return ["SetTestSelection", "Visibility", "ResetVisibility", "AllStems"]

    def _getStockId(self):
        return "find"

    def _getTitle(self):
        return "_Select"

    def getTooltip(self):
        return "Select indicated tests"

    def getTabTitle(self):
        return "Selection"

    def messageBeforePerform(self):
        return "Selecting tests ..."

    def messageAfterPerform(self):
        return "Selected " + self.describeTests() + "."

    # No messageAfterPerform necessary - we update the status bar when the selection changes inside TextTestGUI
    def getFilterList(self, app):
        optionMap = self.optionGroup.getOptionValueMap()
        useTmpFiles = self.optionGroup.getOption("std") and "std" not in optionMap
        return app.getFilterList(self.rootTestSuites, optionMap, useTmpFiles=useTmpFiles)

    def makeNewSelection(self):
        # Get strategy. 0 = discard, 1 = refine, 2 = extend, 3 = exclude
        strategy = self.selectionGroup.getSwitchValue("current_selection")
        return self._makeNewSelection(strategy, includeEmptySuites=False)

    def notifyReset(self, *args):
        self.optionGroup.reset()
        self.selectionGroup.reset()
        self.filteringGroup.reset()

    def _makeNewSelection(self, strategy, includeEmptySuites):
        selectedTests = []
        suitesToTry = self.getSuitesToTry()
        for suite in self.rootTestSuites:
            if suite in suitesToTry:
                filters = self.getFilterList(suite.app)
                reqTests = self.getRequestedTests(suite, filters, strategy, includeEmptySuites)
                newTests = self.combineWithPrevious(reqTests, suite.app, strategy)
            else:
                newTests = self.combineWithPrevious([], suite.app, strategy)

            selectedTests += newTests
        return selectedTests

    def performOnCurrent(self):
        newSelection = self.makeNewSelection()
        criteria = " ".join(self.getCommandLineArgs(self.optionGroup, onlyKeys=self.appKeys))
        self.notify("SetTestSelection", newSelection, criteria,
                    self.selectionGroup.getSwitchValue("select_in_collapsed_suites"))

    def matchesVersions(self, test):
        versionSelection = self.optionGroup.getOptionValue("vs")
        if len(versionSelection) == 0:
            return True
        elif versionSelection == "<default>":
            return len(test.app.versions) == 0
        else:
            versions = set(versionSelection.split("."))
            appVersions = set(test.app.versions + test.app.getBaseVersions())
            return versions.issubset(appVersions)

    def findMainApps(self):
        apps = set(self.validApps)
        for app in self.validApps:
            apps.difference_update(app.extras)
        return apps

    def adjustForExtraVersions(self, rootSuites):
        apps = self.findMainApps()
        if len(apps) == len(rootSuites):
            return rootSuites

        versionSelection = self.optionGroup.getOptionValue("vs")
        removeApps = []
        for app in apps:
            refApp = app.getRefVersionApplication(versionSelection)
            newExtras = set(refApp.getConfigValue("extra_version"))
            for extra in app.extras:
                if len(set(extra.versions).intersection(newExtras)) == 0:
                    removeApps.append(extra)
        return [s for s in rootSuites if s.app not in removeApps]

    def getSuitesToTry(self):
        # If only some of the suites present match the version selection, only consider them.
        # If none of them do, try to filter them all
        toTry = list(filter(self.matchesVersions, self.rootTestSuites))
        if len(toTry) == 0:
            return self.adjustForExtraVersions(self.rootTestSuites)
        else:
            return toTry

    def getRequestedTests(self, suite, filters, strategy, includeEmptySuites):
        self.notify("ActionProgress")  # Just to update gui ...
        if strategy == 1:  # refine, don't check the whole suite
            tests = [test for test in self.currTestSelection if test.app is suite.app and test.isAcceptedByAll(filters)]
        else:
            tests = self.getRequestedTestsFromSuite(suite, filters, includeEmptySuites)

        # Some filters need to look at the selection as a whole to decide what to do
        for testFilter in filters:
            tests = testFilter.refine(tests)
        return tests

    def getRequestedTestsFromSuite(self, suite, filters, includeEmptySuites):
        if not suite.isAcceptedByAll(filters):
            return []
        else:
            if suite.classId() == "test-suite":
                tests = []
                testcases = self.findTestCaseList(suite)
                if len(testcases) or not includeEmptySuites:
                    for subSuite in testcases:
                        self.notify("ActionProgress")  # Just to update gui ...
                        tests += self.getRequestedTestsFromSuite(subSuite, filters, includeEmptySuites)
                    return tests
            # include empty suites here, if they match
            return [suite]

    def combineWithPrevious(self, reqTests, app, strategy):
        # Strategies: 0 - discard, 1 - refine, 2 - extend, 3 - exclude
        # If we want to extend selection, we include test if it was previsouly selected,
        # even if it doesn't fit the current criterion
        if strategy < 2:
            return reqTests
        else:
            extraRequested = [test for test in reqTests if test not in self.currTestSelection]
            if strategy == 2:
                selectedThisApp = [test for test in self.currTestSelection if test.app is app]
                return extraRequested + selectedThisApp
            elif strategy == 3:
                return extraRequested

    def findTestCaseList(self, suite):
        version = self.optionGroup.getOptionValue("vs")
        if len(version) == 0:
            return suite.testcases

        if version == "<default>":
            version = ""

        fullVersion = suite.app.getFullVersion()
        baseVersions = suite.app.getBaseVersions()
        versionToUse = self.findCombinedVersion(version, fullVersion, baseVersions)
        self.selectDiag.info("Trying to get test cases for " + repr(suite) + ", version " + versionToUse)
        return suite.findTestCases(versionToUse)

    def findCombinedVersion(self, version, fullVersion, baseVersions):
        combined = self.getGivenVersion(version, baseVersions)
        if len(fullVersion) > 0 and len(combined) > 0:
            parts = version.split(".")
            for appVer in fullVersion.split("."):
                if appVer not in parts:
                    combined += "." + appVer
        return combined

    def getGivenVersion(self, version, baseVersions):
        parts = [v for v in version.split(".") if v not in baseVersions]
        return ".".join(parts)

    def filterTests(self, *args):
        self.notify("Status", "Filtering tests ...")
        self.notify("ActionStart")
        newSelection = self._makeNewSelection(strategy=0, includeEmptySuites=True)
        strategy = self.filteringGroup.getSwitchValue("current_filtering")
        toShow = self.findTestsToShow(newSelection, strategy)
        signalToSend = "ResetVisibility" if self.dynamic else "Visibility"
        self.notify(signalToSend, toShow, True)
        self.notify("ActionProgress")
        toHide = self.findTestsToHide(newSelection, strategy)
        self.notify(signalToSend, toHide, False)
        self.notify("ActionStop")
        self.notify("Status", "Changed filtering by showing " + str(len(toShow)) +
                    " tests/suites and hiding " + str(len(toHide)) + ".")

    def findTestsToShow(self, newSelection, strategy):
        if strategy == 0 or strategy == 2:
            return newSelection
        else:
            return []

    def findTestsToHide(self, newSelection, strategy):
        if strategy == 0 or strategy == 1:
            return self.findTestsNotIn(newSelection)
        else:
            return []

    def getFilterTooltip(self):
        return "filter tests to show only those indicated"

    def createFilterButton(self):
        return self._createButton(self.filterAction, self.getFilterTooltip())

    def createFrame(self, name, group, button):
        frame = Gtk.Frame.new(name)
        frame.set_label_align(0.5, 0.5)
        frame.set_shadow_type(Gtk.ShadowType.IN)
        frameBox = Gtk.VBox()
        self.fillVBox(frameBox, group)
        self.addCentralButton(frameBox, button, padding=8)
        frame.add(frameBox)
        return frame

    def getNewSwitchName(self, switchName, optionGroup):
        if optionGroup is self.selectionGroup:
            return "Current selection"
        elif optionGroup is self.filteringGroup:
            return "Current filtering"
        else:
            return switchName

    def getNaming(self, switchName, cleanOption, optionGroup):
        return guiplugins.ActionTabGUI.getNaming(self, self.getNewSwitchName(switchName, optionGroup), cleanOption, optionGroup)

    def createButtons(self, vbox):
        selFrame = self.createFrame("Selection", self.selectionGroup, self.createButton())
        vbox.pack_start(selFrame, False, False, 8)
        filterFrame = self.createFrame("Filtering", self.filteringGroup, self.createFilterButton())
        vbox.pack_start(filterFrame, False, False, 8)
        self.addCentralButton(vbox, self.createResetButton(), padding=16)


class HideSelected(guiplugins.ActionGUI, AllTestsHandler):
    def __init__(self, *args):
        guiplugins.ActionGUI.__init__(self, *args)
        AllTestsHandler.__init__(self)

    def _getTitle(self):
        return "Hide selected"

    def messageBeforePerform(self):
        return "Hiding all tests that are currently selected ..."

    def getTooltip(self):
        return "Hide all tests that are currently selected"

    def getSignalsSent(self):
        return ["Visibility"]

    def performOnCurrent(self):
        self.notify("Visibility", self.currTestSelection, False)


class HideUnselected(guiplugins.ActionGUI, AllTestsHandler):
    def __init__(self, *args):
        guiplugins.ActionGUI.__init__(self, *args)
        AllTestsHandler.__init__(self)

    def _getTitle(self):
        return "Show only selected"

    def messageBeforePerform(self):
        return "Showing only tests that are currently selected ..."

    def getTooltip(self):
        return "Show only tests that are currently selected"

    def getSignalsSent(self):
        return ["Visibility"]

    def performOnCurrent(self):
        self.notify("Visibility", self.findTestsNotIn(self.currTestSelection), False)


class ShowAll(guiplugins.BasicActionGUI, AllTestsHandler):
    def __init__(self, allApps, dynamic, *args):
        guiplugins.BasicActionGUI.__init__(self, allApps, dynamic, *args)
        AllTestsHandler.__init__(self)
        self.dynamic = dynamic

    def _getTitle(self):
        return "Show all"

    def messageBeforePerform(self):
        return "Showing all tests..."

    def getTooltip(self):
        return "Show all tests"

    def getSignalsSent(self):
        return ["Visibility", "ResetVisibility"]

    def performOnCurrent(self):
        signalToSend = "ResetVisibility" if self.dynamic else "Visibility"
        self.notify(signalToSend, self.findAllTests(), True)


class SaveSelection(guiplugins.ActionDialogGUI):
    def __init__(self, allApps, dynamic, *args):
        guiplugins.ActionDialogGUI.__init__(self, allApps, dynamic, *args)
        possibleDirs = self.getFilterFileDirs(allApps, useOwnTmpDir=False)
        self.addOption("f", "enter filter-file name =", possibleDirs=possibleDirs, saveFile=True)
        if not dynamic:
            # In the static GUI case, we also want radiobuttons specifying
            # whether we want to save the actual tests, or the selection criteria.
            self.addSwitch("tests", "Save", options=[
                           "_List of selected tests", "C_riteria entered in the Selection tab\n(Might not match current selection, if it has been modified)"])
        self.selectionCriteria = ""
        self.dynamic = dynamic
        self.rootTestSuites = []

    def addSuites(self, suites):
        self.rootTestSuites += suites

    def _getStockId(self):
        return "save-as"

    def _getTitle(self):
        return "S_ave Selection..."

    def getTooltip(self):
        return "Save selected tests in file"

    def getSignalsSent(self):
        return ["WriteTestsIfSelected"]

    def writeTestList(self, file):
        file.write("-tp ")
        for suite in self.rootTestSuites:
            versionSuffix = suite.app.versionSuffix()
            if "copy_" not in versionSuffix:  # Don't save copy names in selections, causes confusion
                file.write("appdata=" + suite.app.name + versionSuffix + "\n")
            self.notify("WriteTestsIfSelected", suite, file)

    def notifySetTestSelection(self, dummy, criteria="", *args):
        self.selectionCriteria = criteria

    def getConfirmationMessage(self):
        fileName = self.optionGroup.getOptionValue("f")
        if fileName and os.path.isfile(fileName):
            return "\nThe file \n" + fileName + "\nalready exists.\n\nDo you want to overwrite it?\n"

    def getConfirmationDialogSettings(self):
        return Gtk.STOCK_DIALOG_QUESTION, "Query"

    def notifySaveSelection(self, fileName, writeCriteria=False):
        try:
            file = open(fileName, "w")
            if writeCriteria:
                file.write(self.selectionCriteria + "\n")
            else:
                self.writeTestList(file)
            file.close()
        except IOError as e:
            raise plugins.TextTestError("\nFailed to save selection:\n" + str(e) + "\n")

    def getFileName(self):
        fileName = self.optionGroup.getOptionValue("f")
        if not fileName:
            raise plugins.TextTestError("Cannot save selection - no file name specified")
        elif os.path.isdir(fileName):
            raise plugins.TextTestError("Cannot save selection - existing directory specified")
        else:
            return fileName

    def performOnCurrent(self):
        fileName = self.getFileName()
        writeCriteria = not self.dynamic and self.optionGroup.getSwitchValue("tests")
        self.notifySaveSelection(fileName, writeCriteria)

    def messageAfterPerform(self):
        return "Saved " + self.describeTests() + " in file '" + self.optionGroup.getOptionValue("f") + "'."


class LoadSelection(guiplugins.ActionDialogGUI):
    def __init__(self, allApps, *args):
        guiplugins.ActionDialogGUI.__init__(self, allApps, *args)
        possibleDirs = self.getFilterFileDirs(allApps, useOwnTmpDir=True)
        self.addOption("f", "select filter-file", possibleDirs=possibleDirs, selectFile=True)
        self.rootTestSuites = []

    def addSuites(self, suites):
        self.rootTestSuites += suites

    def isActiveOnCurrent(self, *args):
        return True

    def getSignalsSent(self):
        return ["SetTestSelection"]

    def _getStockId(self):
        return "open"

    def _getTitle(self):
        return "_Load Selection..."

    def getTooltip(self):
        return "Load test selection from file"

    def performOnCurrent(self):
        fileName = self.optionGroup.getOptionValue("f")
        if fileName:
            newSelection = self.makeNewSelection(fileName)
            self.notify("SetTestSelection", newSelection, "-f " + fileName, True)
            self.notify("Status", "Loaded test selection from file '" + fileName + "'.")
        else:
            self.notify("Status", "No test selection loaded.")

    def makeNewSelection(self, fileName):
        tests = []
        for suite in self.rootTestSuites:
            filters = suite.app.getFiltersFromFile(fileName, self.rootTestSuites)
            self.notify("ActionProgress")
            tests += suite.testCaseList(filters)
        return tests

    def getSizeAsWindowFraction(self):
        # size of the dialog
        return 0.6, 0.8

    def messageBeforePerform(self):
        return "Loading test selection ..."

    def messageAfterPerform(self):
        pass


class SelectInSameGroup(guiplugins.ActionGUI):
    def getSignalsSent(self):
        return ["SelectInGroup"]

    def _getTitle(self):
        return "Select Tests With Same Diff"

    def performOnCurrent(self):
        self.notify("SelectInGroup", self.currFileSelection[0][1])

    def isActiveOnCurrent(self, *args):
        if not guiplugins.ActionGUI.isActiveOnCurrent(self) or len(self.currFileSelection) != 1:
            return False

        comparison = self.currFileSelection[0][1]
        return comparison and not comparison.hasSucceeded()


def getInteractiveActionClasses(dynamic):
    classes = [SaveSelection, SelectTests, HideUnselected, HideSelected, ShowAll, LoadSelection]
    if dynamic:
        classes.append(SelectInSameGroup)
    return classes
